/* Copyright 2021 Kevin Ryde

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   This file is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this file.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* ------------------------------------------------------------------------- */
/* Kevin Ryde, August 2021

   Usage: ./a.out [-v] n1 n2 nlo-nhi

   Calculate terms of A347414 by brute force: iterate through all rooted
   forests of n vertices, find their automorphism orbits with Nauty, and
   count how many distinct partitions of n occur among them.  This is
   suitable only for small n.  Created for use with Nauty 2.7r1.

   The default is to attempt all n from 1 to 32.  Individual n or ranges can
   be given on the command line

       ./a.out 2 5-7 13

   Output is like

       n=1  a(n) = 1
       n=2  a(n) = 2
       n=3  a(n) = 3
       n=4  a(n) = 5
       n=5  a(n) = 6

   Command-line option "-v" prints some more verbose information, including
   progress messages with "\r" carriage return overwriting designed for
   output to a terminal.

   Compiling
   ---------

   Put this file in a Nauty build directory which has had its "./configure"
   and "make nauty1.a" done, then compile to an executable a.out,

       gcc -O2 a347414-nauty.c nauty1.a

   If Nauty is installed in system locations (such as Debian or similar),
   then #include "nauty.h" must become #include <nauty/nauty.h> and

       gcc -O2 a347414-nauty.c -lnauty1

   The code uses 1-word graphs so the nauty1.a library suffices and is a
   little faster than the full nauty.a.  The number of vertices is limited
   to n <= WORDSIZE, where the Nauty WORDSIZE is either 32 or 64.  More than
   n=32 is expected to be too slow anyway.  (To attempt more, increase
   SEEN_TABLE_SIZE for more partitions and if on a 32-bit machine then
   compile -DWORDSIZE=64 and link nautyL1.a instead of nauty1.a.)

   Notes
   -----

   A rooted forest is passed to Nauty as an undirected graph with the roots
   distinguished by an initial vertex partition.  This seems faster than a
   directed graph with edges from child to parent (or vice versa).

   Nauty's densenauty() returns automorphism orbits in the form of an array
   orbits[v] = smallest vertex number in the orbit of v.  This is turned
   into orbit sizes here, sorted, and made into a bitwise "partition_code"
   ready to bsearch() in the seen_table[].

   The number of partitions of n grows much slower than the number of rooted
   forests so that seen_table[] size is modest.  Not all partitions occur,
   but enough of them do that it doesn't seem worth any fancier data
   structure than just the complete list and mark those seen.
*/

/* Change this to WANT_ASSERTS 1 to run the assert() statements for various
   consistency checks.  */
#define WANT_ASSERTS  0

/* Whether include \r progress messages in -v option. */
#define WANT_PROGRESS  1

/* Change this to WANT_DEBUG 1 for mucho debug prints. */
#define WANT_DEBUG  0


/* The code here is 1-word graph storage and operations so MAXN accordingly.
   Do this before nauty.h so as to get its 1-word versions of macros.
*/
#define MAXN  WORDSIZE

#if ! WANT_ASSERTS
#define NDEBUG         /* for <assert.h> */
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include "nauty.h"

#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

int option_verbose = 0;

#define MIN(x,y)  ((x)<(y) ? (x) : (y))


/* ------------------------------------------------------------------------- */
/* seen_table[] is a sorted list of all possible partition_code values and a
   "seen" flag marking those seen as orbit sizes.

   SEEN_TABLE_SIZE must be big enough for all partitions of n, which means
   A000041(n) or more, for whatever maximum n will be attempted.
*/

/* this size is enough up to numbpart(32) == 8349 partitions of n=32 */
#define SEEN_TABLE_SIZE  10000

/* num_partitions[n] = A000041(n) = how many partitions of n */
const int num_partitions[] = {
  1, 1, 2, 3, 5, 7, 11, 15, 22, 30, 42, 56, 77, 101, 135, 176, 231, 297,
  385, 490, 627, 792, 1002, 1255, 1575, 1958, 2436, 3010, 3718, 4565,
  5604, 6842, 8349, 10143, 12310, 14883, 17977, 21637, 26015, 31185,
  37338, 44583, 53174, 63261, 75175, 89134, 105558, 124754, 147273,
  173525
};
#define NUM_PARTITIONS_LEN  (sizeof(num_partitions)/sizeof(num_partitions[0]))

/* partition_code is 1-bits marking the start of each part of the partition,
   as per A225620.  Parts are sorted in ascending order and coded from most
   to least significant bit.  Eg. a partition of n=12,

       high         low
       1 1 10 1000 1000   partition_code (in binary)
       ^ ^ \/ \--/ \--/
       1 1  2   4    4       1+1+2+4+4 = 12

   This coding needs at most MAXN bits so Nauty type "setword".

   init_seen() (and init_seen_recurse()) make all partitions in increasing
   order of partition_code.  The generating method is of no great
   sophistication and in particular is not efficient at noticing when no
   bigger part can be added.  But its work is a tiny part of the total.
*/

struct seen {
  setword partition_code;
  char seen;
};
struct seen  seen_table[SEEN_TABLE_SIZE];
int seen_table_len;


/* init_seen_recurse() takes a partition_code so far and
   pos = next place to consider setting a 1-bit.
   t = smallest next part to consider adding.
   Adding a part is by set 1-bit at pos, decrease pos by t.
*/
void
init_seen_recurse (setword partition_code, int pos, int t)
{
  DEBUG(printf("init_seen_recurse() partition_code %#X pos=%d t=%d\n",
               partition_code, pos, t));
  assert (t >= 1);
  assert (pos < MAXN);  /* bit positions are pos=MAXN-1 downwards */
  assert (pos >= -1);   /* -1 is fully filled, bigger is partly filled */

  if (pos == -1) {
    /* partition_code is fully filled, so a complete partition */
    if (seen_table_len >= SEEN_TABLE_SIZE) {
      fprintf(stderr, "oops, exceeded SEEN_TABLE_SIZE\n");
      exit(1);
    }

    /* ascending order is vital for correct bsearch(), so check always */
    if (seen_table_len >= 1
        && partition_code <= seen_table[seen_table_len-1].partition_code) {
      fprintf(stderr, "oops, partition_code meant to be ascending order\n");
      exit(1);
    }

    seen_table[seen_table_len].partition_code = partition_code;
    seen_table[seen_table_len].seen           = 0;
    seen_table_len++;
    return;
  }
  if (pos - t < -1) {
    /* t is too big for the bits pos..0 inclusive left in partition_code,
       and any bigger t would be too, so nothing more
    */
    return;
  }

  /* cases with no parts equal to t, only parts bigger than t */
  init_seen_recurse(partition_code, pos, t+1);

  /* cases with one or more parts equal to t, then further bigger parts */
  while(pos-t >= -1) {
    partition_code += 1<<pos;
    pos -= t;
    init_seen_recurse(partition_code, pos, t+1);
  }
}
void
init_seen (int n)
{
  assert (1 <= n && n <= MAXN);
  seen_table_len = 0;
  init_seen_recurse(0, n-1, 1);
  if (n < NUM_PARTITIONS_LEN && seen_table_len != num_partitions[n]) {
    fprintf(stderr, "oops, seen_table_len != num_partitions\n");
    exit(1);
  }
}

/* xp and yp are pointers to "struct seen".
   Return 1,0,-1 according as order of their partition_code values.
*/
int
seen_cmp_partition_code (const void *xp, const void *yp)
{
  /* not just "x-y", because setword might be bigger than int */
  setword x = ((struct seen *) xp)->partition_code;
  setword y = ((struct seen *) yp)->partition_code;
  return (x > y ? 1 : x==y ? 0 : -1);
}

/* locate partition_code within seen_table[] and mark it as seen */
void
add_seen_partition_code (setword partition_code)
{
  struct seen key, *found;
  DEBUG(printf("add_orbitsizes_partition_code() partition_code %#X\n",
               partition_code));

  key.partition_code = partition_code;
  found = bsearch (&key, seen_table, seen_table_len, sizeof(seen_table[0]),
                   seen_cmp_partition_code);
  assert(found != NULL);
  found->seen = 1;
}

int
partition_to_partition_code (const int *array, int len)
{
  setword  partition_code = 0;
  int i;
  DEBUG(printf("partition_to_partition_code() len=%d\n", len));
  assert(len>=0);
#if WANT_ASSERTS
  { int total = 0;
    for (i = 0; i < len; i++) {
      assert (array[i] >= 1);  /* all parts >=1 */
      total += array[i];
    }
    assert(total <= MAXN);     /* total must fit in setword */
  }
#endif
  for (i = 0; i < len; i++)
    partition_code = ((partition_code << 1) + 1) << (array[i]-1);
  return(partition_code);
}

/* orbitsizes[0 .. num_orbits-1] is a partition of n, so that sum
   orbitsizes[0] + ... + orbitsizes[num_orbits-1] == n, with terms in
   ascending order.
*/
void
add_seen_orbitsizes (const int *orbitsizes, int num_orbits)
{
  add_seen_partition_code
    (partition_to_partition_code(orbitsizes,num_orbits));
}

/* ------------------------------------------------------------------------- */
/* Rooted forests are iterated here in the manner of Beyer and Hedetniemi,

       Terry Beyer and Sandra Mitchell Hedetniemi, "Constant Time Generation
       of Rooted Trees", SIAM Journal of Computing, volume 9, number 4,
       November 1980, pages 706-712.

   Here a rooted forest is vertices v = 0..n-1 and is represented by depths
   vector depths[0..n-1] with the roots at depth 0, their children are depth
   1, and so on.

   (See A346913 for Beyer and Hedetniemi's iteration.  As compared to there,
   here discards the tree root so as to leave the rest as a forest, and what
   would then be level=2 in A346913 for the forest roots becomes depth=0.)
*/

/* Set depths[0..n-1] to the first rooted forest of n vertices, which means
   0,1,2,...,n-1 path-n downward.
*/
void
depths_first (int *depths, int n)
{
  int i;
  for (i=0; i<n; i++) depths[i] = i;
}

/* depths[0..n-1] is a rooted forest of n vertices.
   If the next forest exists, then change depths[] to it and return 1.
   If no next forest exists, then return 0.
*/
int
depths_next (int *depths, int n)
{
  int p;
  DEBUG(printf("  depths_next()\n"));
  assert (n<1 || depths[0] == 0);

  for (p = n-1; p>0; p--) {    /* find p end-most with depths[p] > 0 */
    int d = depths[p];
    if (d > 0) {
      int q;
      for (q = p-1; q>=0; q--) {    /* find q = parent of p */
        assert (q >= 0);
        if (depths[q] < d) {
          DEBUG(printf("    p=%d (depth d=%d) q=%d  block len=%d\n", p,d,q,p-q));

          /* Replace depths[p..n-1] with copies of depths[q..p-1],
             inclusing a partial copy at the end if necessary
             (if p..n-1 is not a multiple of length q..p-1)
          */
          while (p < n)
            depths[p++] = depths[q++];
          return(1);
        }
      }
      assert(0);  /* oops, no parent of p */
    }
  }

  /* no next forest exists */
  return(0);
}


/* ------------------------------------------------------------------------- */
/* Progress messages are printed roughly every PROGRESS_INTERVAL seconds.
   Progress percentage is based how far through num_rooted_forests[n].

   The test for whether to print the status is based on counter
   progress_remaining, so as to avoid the slowdown of say a system call
   to time() for every forest.
*/

/* A000081 */
const double num_rooted_forests[] = {
  1.0, 1.0, 2.0, 4.0, 9.0, 20.0, 48.0, 115.0, 286.0, 719.0,
  1842.0, 4766.0, 12486.0, 32973.0, 87811.0, 235381.0, 634847.0,
  1721159.0, 4688676.0, 12826228.0, 35221832.0, 97055181.0,
  268282855.0, 743724984.0, 2067174645.0, 5759636510.0,
  16083734329.0, 45007066269.0, 126186554308.0, 354426847597.0
};
#define NUM_ROOTED_FORESTS_LEN \
  (sizeof(num_rooted_forests) / sizeof(num_rooted_forests[0]))
#define PROGRESS_INTERVAL  5  /* seconds between prints */
time_t progress_lasttime;
int progress_block = 100;  /* how many forests between progress prints */
int progress_remaining;    /* how many remaining until next print */
int progress_printed = 0;

#define PROGRESS(n,forests_count)                                       \
  do {                                                                  \
    if (WANT_PROGRESS && option_verbose && --progress_remaining < 0)    \
      progress_print(n,forests_count);                                  \
  } while (0)

void
progress_print (int n, double forests_count)
{
  progress_printed = 1;
  if (n < NUM_ROOTED_FORESTS_LEN && num_rooted_forests[n])
    printf("n=%d, %.0lf forests, %.1lf%%\r",
           n, forests_count,
           100.0 * forests_count / num_rooted_forests[n]);
  else
    printf("n=%d, %.0lf forests\r", n, forests_count);

  {
    /* The previous progress_block many forests took elapsed_time to run.
       Recalculate progress_block to be about 5 seconds worth of forests.
       But try to beware of clock jumps or process stops.
     */
    time_t t = time(NULL);
    double elapsed_time = labs(t - progress_lasttime);
    if (elapsed_time) {
      int per_second = progress_block / elapsed_time;
      progress_block = MIN(PROGRESS_INTERVAL*per_second+1,
                           2*progress_block+1);
    } else {
      /* 0 seconds for the last block, try doubling */
      progress_block = 2*progress_block + 1;
    }
    if (progress_block < 0) progress_block = 100;
    progress_remaining = progress_block;
    progress_lasttime = t;
  }
}

/* ------------------------------------------------------------------------- */

/* depths[0..n-1] is a rooted forest as pre-order depths.
   Convert it to an undirected Nauty graph g,m,n.
   Undirected means arcs both ways for edges.
   g must have space for n vertices.
   Any existing contents in g are replaced.
*/
void
depths_to_graph (const int *depths, graph *g, int m, int n)
{
  int prev[MAXN];   /* prev[d] = vertex number last seen at depth d */
  int v;
  assert (n >= 0);
  assert (n <= MAXN);
  assert (m >= SETWORDSNEEDED(n));

  EMPTYGRAPH(g, m,n);
  for (v = 0; v < n; v++) {
    int d = depths[v];
    assert(d >= 0);
    assert(d < n);
    if (d > 0) {
      int to = prev[d-1];   /* parent of v */
      ADDONEEDGE(g, v,to, m);
    }
    prev[d] = v;
  }
}

int
depths_to_num_roots (const int *depths, int n)
{
  int v, num_roots = 0;
  for (v=0; v<n; v++)
    num_roots += (depths[v]==0);
  return (num_roots);
}

/* depths[0..n-1] is a rooted forest as pre-order depths.  Set lab[] and
   ptn[] to an initial partition which distinguishes the roots for Nauty.

   lab[] lists the root vertex numbers first (those with depths[v]==0),
   followed by all other vertex numbers.  ptn[] is 1 everywhere except 0
   at the last root and 0 at the last of the others (which is v=n-1) to
   mark the ends of the partitions.

   As a note, it'd be straightforward to make an initial partition with all
   vertices by depths, so all v with depths[v]==0 separate, then all
   depths[v]==1 separate, and so on.  But doing that seemed no measurable
   speedup over distinguishing the roots and letting Nauty do the rest.
*/
void
depths_to_lab_ptn (const int *depths, int n, int *lab, int *ptn)
{
  int num_roots = depths_to_num_roots(depths,n);
  {
    int rpos = 0;          /* roots store to rpos */
    int opos = num_roots;  /* others store to opos */
    int v;
    for (v = 0; v < n; v++) {
      if (depths[v]) lab[opos++] = v;
      else           lab[rpos++] = v;
    }
    assert(opos == n);
    assert(rpos == num_roots);
  }
  {
    int v;
    for (v = 0; v < n; v++) ptn[v] = 1;
    assert(num_roots-1 >= 0);
    assert(n-1 >= 0);
    ptn[num_roots-1] = 0;   /* mark last of the roots */
    ptn[n-1] = 0;           /* mark last of the others */
  }
}

/* xp and yp are pointers to ints */
int
orbitsize_cmp_ascending (const void *xp, const void *yp)
{
  /* orbit sizes are small enough to allow subtract "x-y" */
  int x = (int) * (const int *) xp;
  int y = (int) * (const int *) yp;
  return (x - y);
}

static DEFAULTOPTIONS_GRAPH(options);

void
process_depths (const int *depths, int n)
{
  const int m = SETWORDSNEEDED(MAXN);
  setword g[MAXN*SETWORDSNEEDED(MAXN)];
  setword g_out[MAXN*SETWORDSNEEDED(MAXN)];
  int lab[MAXN];
  int ptn[MAXN];
  int orbits[MAXN];
  int orbitsizes[MAXN];
  int v;
  statsblk stats;

  DEBUG(printf("process_depths() n=%d\n", n));
  assert(n <= MAXN);                  /* g[] array size */
  assert(m <= SETWORDSNEEDED(MAXN));  /* g[] array size */
  assert(m >= SETWORDSNEEDED(n));
  DEBUG(printf("  depths ");
        for(v=0; v<n; v++) printf(" %d", depths[v]);
        printf("\n"));

  depths_to_graph(depths, g,m,n);
  depths_to_lab_ptn (depths,n, lab,ptn);
  densenauty ((graph*)g, lab, ptn, orbits, &options, &stats, m,n, g_out);

  memset (orbitsizes, '\0', n*sizeof(orbitsizes[0]));
  for (v = 0; v < n; v++) {
    int u = orbits[v];  /* smallest vertex in the orbit of v */
    assert (0 <= u && u < n);
    orbitsizes[u]++;    /* how many in the orbit of u */
  }

  DEBUG(printf("    orbitsizes ");
        for(v=0; v<n; v++) printf(" %d", (int) orbitsizes[v]);
        printf("\n"));

  qsort (orbitsizes, n, sizeof(orbitsizes[0]), orbitsize_cmp_ascending);

  DEBUG(printf("    sorted     ");
        for(v=0; v<n; v++) printf(" %d", (int) orbitsizes[v]);
        printf("\n"));
#if WANT_ASSERTS
  /* first n - stats.numorbits entries of orbitsizes[] = 0, rest != 0 */
  for(v=0; v<n; v++)
    assert( (orbitsizes[v]==0) == (v < n - stats.numorbits));
#endif

  /* after initial 0 entries in orbitsizes[] */
  add_seen_orbitsizes(orbitsizes + n - stats.numorbits,
                      stats.numorbits);
}

/* calculate and print A347414(n) */
void
calculate_n (int n)
{
  double forests_count = 0;
  int depths[MAXN];
  int num_seen;

  if (! (1 <= n && n <= MAXN)) {
    fprintf(stderr, "n out of range, must be 1 to %d, got %d\n", MAXN, n);
    exit(1);
  }
  {
    int m = SETWORDSNEEDED(n);
    nauty_check(WORDSIZE,m,n,NAUTYVERSIONID);
  }
  if (option_verbose) {
    printf("run n=%d\n", n);
  }

  init_seen(n);
  depths_first(depths,n);
  do {
    forests_count++;
    PROGRESS(n,forests_count);
    process_depths(depths,n);
  } while (depths_next(depths,n));

  num_seen = 0;
  { int i;
    for (i=0; i < seen_table_len; i++)
      num_seen += seen_table[i].seen;
  }

  if (option_verbose) {
    printf("n=%d end, rooted forests count %.0lf, a(%d) = %d\n",
           n, forests_count, n, num_seen);
    printf("\n");

  } else {
    printf("n=%d  a(n) = %d%s\n", n, num_seen,
           progress_printed ? "                             " : "");
  }
  if (n < NUM_ROOTED_FORESTS_LEN)
    assert (forests_count == num_rooted_forests[n]);
}

int
main (int argc, char **argv)
{
  int seen_n = 0;
  int i;
  setbuf(stdout,NULL);
  options.defaultptn = 0;   /* lab,ptn will be given */

  for (i = 1; i < argc; i++) {
    const char *arg = argv[i];
    int n, len;
    unsigned lo, hi;
    if (strcmp(arg,"-v") == 0) {
      option_verbose = 1;
      printf("NAUTYVERSION is %s\n", NAUTYVERSION);
      printf("SEEN_TABLE_SIZE %u is %u bytes\n", SEEN_TABLE_SIZE, sizeof(seen_table));
      printf("\n");
    } else if (sscanf(arg, "%d%n", &n, &len) == 1
               && len == strlen(arg)) {
      seen_n = 1;
      calculate_n(n);
    } else if (sscanf(arg, "%u-%u%n", &lo, &hi, &len) == 2
               && len == strlen(arg)) {
      seen_n = 1;
      for (n = lo; n <= hi; n++)
        calculate_n(n);
    } else {
      fprintf(stderr, "Unrecognised command line option: %s\n", arg);
      exit(1);
    }
  }

  if (!seen_n) {
    int n;
    for (n = 1; n <= MAXN; n++) {
      calculate_n(n);
    }
  }
  return 0;
}
