\\ Copyright 2020, 2021, 2022 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0);
default(strictargs,1);

option_verbose=0;
option_progress=1;

read("../vpar.gp");
read("../vpar-wip.gp");
read("../test-helpers.gp");
read("OEIS-data.gp");
read("OEIS-data-wip.gp");



\\-----------------------------------------------------------------------------
{
  my(v=OEIS_data("A347414"));
  check_equal(vector(#v,n, lex(numbpart(n), v[n])),
              vector(#v,n, if(n==5 || n>=7, 1, 0)),
              "all partitions until n=5 and n>=7");
}

\\ return vector of partitions occurring
A347414_partitions(n) =
{
  my(m=Map(),
     vpar=vpar_unlabelled_first(n));
  until(!vpar_unlabelled_nextR(~vpar,1),  \\ forests
        mapput(m, Vecsmall(vecsort(vpar_orbit_sizes(vpar))), 1));
  Vec(Mat(m)[,1]);
}
A347414_by_search(n) = #A347414_partitions(n);

{
  \\ partitions which don't occur
  if(0,
     for(n=8,12,
        print("n=",n);
        my(all=partitions(n),
           got=A347414_partitions(n),
           non=setminus(all,got));
        foreach(non,p, print(Vec(p))));
     quit);
}

\\-----------------------------------------------------------------------------
\\ example n=7

{
  my(n=7,
     P=A347414_partitions(n));
  check_equal(#P,          13);
  check_equal(numbpart(n), 15);

  check_equal(setminus(P,partitions(n)), []);
  check_equal(setminus(partitions(n), P),
              [Vecsmall([2,5]), Vecsmall([3,4])],
              "n=7 not occurring");
}

{
  \\ smaller one
  my(n=5,
     P=A347414_partitions(n));
  check_equal(#P,          6);
  check_equal(numbpart(n), 7);

  check_equal(setminus(P,partitions(n)), [  ]);
  check_equal(setminus(partitions(n),P), [ Vecsmall([2,3]) ]);
}


\\-----------------------------------------------------------------------------
\\ n=9 example

\\   roots:     a      d        orbits:  a   b   c   d   e   f
\\              |     / \       sizes:   1 + 1 + 2 + 1 + 2 + 2 = 9
\\   children:  b    e   e
\\             / \   |   |
\\            c   c  f   f

{
  my(n=9,
     \\    1 2 3 4  5 6 7 8 9
     vpar=[0,1,2,2, 0,5,5,6,7],
     p=[1,1,1, 2,2,2]);
  check_equal(#vpar, n);
  check_equal(vpar_orbit_sets(vpar), [[1],[2],[3,4], [5], [6,7], [8,9]]);
  check_equal(vpar_orbit_sizes(vpar), [1,1,2, 1,2,2]);
  check_equal(vecsort(vpar_orbit_sizes(vpar)), p);

  \\ another way
  \\       1 2 3  4 5  6 7
  my(vpar=[0,1,2, 3,3, 4,5, 6,7]);
  check_equal(vpar_orbit_sizes(vpar), p);
  check_equal(vpar_orbit_sizes(vpar), vecsort([1,1,2, 1,2,2]));

  my(vpar=vpar_unlabelled_first(n),
     count=0);
  until(!vpar_unlabelled_nextR(~vpar,1),  \\ forests
        if(vecsort(vpar_orbit_sizes(vpar))==p,
           \\ print(vpar);
           count++));
  check_equal(count,39,
             "n=9 example total 39 ways");
}


\\-----------------------------------------------------------------------------

partition_to_code(v) =
{
  v==vecsort(v) || error("partition_to_code() expects sorted terms");
  my(P=0);
  for(i=1,#v,
     v[i] >= 1 || error();
     P = ((P<<1)+1) << (v[i]-1));
  P;
}
check_equal(test_to_binary(partition_to_code([1,1,2])), 1110);

\\ like in a347414-nauty.c
init_seen_recurse(P,pos,t) =
{
  \\ print("recurse_partitions P=",test_to_binary(P)," pos="pos" t="t);
  t>=1 || error();
  if(pos==-1,
     listput(partitions_out,P);  \\ new partition_code
     return);

  if(pos - t < -1,
     return); \\ t is too big for the bits pos..0 remaining

  \\ cases with no parts equal to t, only parts bigger than t
  init_seen_recurse(P,pos,t+1);

  \\ cases with one or more parts equal to t, then further bigger parts
  \\ print("recurse_partitions P=",test_to_binary(P)," pos="pos" t="t"  one or more");
  while(pos-t >= -1,
        P += 1<<pos;
        pos -= t;
        init_seen_recurse(P,pos,t+1));
}

\\ set global "partitions_out" to a vector of all partition codes of n
init_seen(n) =
{
  partitions_out=List([]);
  init_seen_recurse(0,n-1,1);
  partitions_out = Vec(partitions_out);
}

{
  all=[]; \\ vector all partition_code to the search limit in one long vector
  for(n=1,32,
     init_seen(n);  \\ make partitions_out vector of partition_codes

     \\ same using partitions() encoded
     by_partitions=vecsort(apply(partition_to_code,partitions(n)));
     if(0 && n<=6,
        print(apply(test_to_binary,partitions_out));
        print(apply(test_to_binary,by_partitions)));

     check_equal(partitions_out, by_partitions,
                 "partitions_out vs by_partitions");
     check_equal(#partitions_out, numbpart(n));
     check_equal(partitions_out, Set(partitions_out));
     all=concat(all,partitions_out);
  );

  check_equal(all,Set(all), "all partitions has no duplicates");
  check_equal(numbpart(32), 8349);

  \\ A225620 is the partition codes, which is all[]
  my(want=OEIS_data("A225620"));
  check_equal(all[1..#want], want,
              "init_seen() = A225620");
}


\\------------------------------------------------------------------------------
\\ A347414 terms by vpar search

{
  if(1,
     my(v=OEIS_data("A347414"));
     for(n=1,#v,
        printf("n=%2d  %4d / %4d = %s%.4f\n",
               n, v[n], numbpart(n),
               if(n%2,"", ""),
               v[n] / numbpart(n)));
     \\ quit;
    );
}
{
  my(v=OEIS_data("A347414"));
  my(limit=13);
  for(n=1,limit,
     check_equal(A347414_by_search(n),
                 v[n],
                 Str("A347414_by_search() vs A347414 data n=",n)));
}


\\------------------------------------------------------------------------------
check_equal(count,'count);
check_equal(none,'none);
check_equal(vpar,'vpar);
print("end");
