\\ Copyright 2022 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

\\-----------------------------------------------------------------------------
\\ Kevin Ryde, July 2022
\\
\\ This is some PARI/GP code to calculate an individual term of A354970 by
\\ double-and-step powering.  It suits large and small n.

default(recover,0);
default(strictargs,1);

{
  my(b2(n)     = (sqr(n)-n)>>1);   \\ binomial(n,2)
  my(b2plus(n) = (sqr(n)+n)>>1);   \\ binomial(n+1,2)

a(n) =
  my(x=0,    \\ a(m), staring at m=0
     y=1,    \\ a(m+1)
     v=valuation(n,2));
  forstep(i=logint(n,2),v+1,-1,  \\ bits high to low
     my(t = b2(y) + 1);
     if(bittest(n,i),
        x+=t;               \\ a(2m+1)
        y+=t;               \\ a(2m+2)
        ,
        y=t+x;              \\ a(2m+1)
        x=b2plus(x) + 1));  \\ a(2m)

  x += b2(y) + 1;         \\ a(2m+1), the 1-bit at v
  for(i=1,v,
     x = b2plus(x) + 1);  \\ a(2m) final doublings
  x;
}

{
  print1("A354970 = ");
  for(n=1,12, print1(a(n),", "));
  print("...");
}

\\ The formulas used from the sequence are
{
  for(n=1,100,
     a(2*n)   == binomial(a(n)+1, 2) + 1         || error();
     a(2*n+1) == binomial(a(n+1), 2) + a(n) + 1  || error();
     a(2*n+2) == binomial(a(n+1)+1, 2) + 1       || error());
}
\\ They allow a pair of terms n,n+1 to double or step by
\\                   for new 0-bit
\\     n     -> 2n      \
\\     n,n+1 -> 2n+1    /     \
\\       n+1 -> 2n+2          /
\\                         for new 1-bit
\\
\\ In the code, at the start of the loop x=a(m) and y=a(m+1) are for
\\ m = n >> (i+1) which is the bits of n above and not including bit
\\ position i (where bit position 0 is the least significant bit).
\\
\\ x=a(0)=0 and y=a(1)=1 initial values are for m=0, which is no bits above
\\ i yet.  a(0)=0 is not part of the sequence but follows from the a(2n+1)
\\ formula at n=0,
\\
\\      a(1)=1             a(1)=1
\\     a(2*0+1) = binomial(a(0+1), 2) + a(0) + 1
\\         1    =     0               + a(0) + 1
\\
\\     so a(0) = 0
\\
\\ The first loop incorporates each bit i from n down to but not including
\\ the least significant 1-bit.  That bit is the explicit b2 after the loop,
\\ to calculate x only.
\\
\\ The second loop is doubling-only to calculate x only for the trailing
\\ 0-bits of n.
\\
\\     n = binary 1..any.. 1 00..00
\\                \------/ v \----/
\\                 first     second
\\                 loop       loop
\\
\\ "v" in the code is the bit position of the least significant 1-bit.
\\
\\ If n is a power of 2 then the first loop does not run, just the v and the
\\ second loop doublings.  This is a complete binary tree, giving A006894.
\\
\\ b2() and b2plus() are binomials and are triangular numbers.
\\ sqr() is used rather than binomial() since in PARI circa verison 2.13,
\\ the sqr() way meansures about 1.4x faster than binomial() on big numbers.
\\ (Perhaps this iis the algorithmic advantage of squaring over
\\ multiplying.)

print("end");
