\\ Copyright 2022 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

\\-----------------------------------------------------------------------------
\\ Kevin Ryde, July 2022
\\
\\ Usage: gp <a356121.gp
\\
\\ This is some PARI/GP code to calculate terms of A356121, the
\\ Matula-Goebel tree number of the rooted binary tree of Colijn-Plazzotta
\\ tree number n.
\\
\\ Requires PARI/GP 2.13 or higher for sqrtint() remainder.

default(recover,0);
default(strictargs,1);

{
  \\ return a 2-element vector [x,y], for n>=2
  my(children(n) =
     my(r, s=sqrtint((n-1)<<1,&r));
     if(r>s,  r-=s;s++, r+=s);
     [s, r>>1]);

  a(n) = if(n==1,1,
            my(x,y); [x,y]=children(n);
            prime(self()(x)) * prime(self()(y)));
}
{
  print1("A356121 = ");
  for(n=1,12, print1(a(n),", "));
  print("...");
}

\\ a(n) is a simple recursive implementation of the sequence formula
\\
\\     a(n) = prime(a(x)) * prime(a(y))
\\
\\ where [x,y] = children(n) are the child subtrees of the root,
\\
\\     x = A002024(n-1)
\\     y = A002260(n-1)
\\
\\ and which are calculated together by sqrtint() and its "remainder".
\\
\\ Recursion suffices since the feasible tree numbers are limited by ever
\\ larger prime(i) for the Matula-Goebel numbers.
\\
\\ prime(i) size varies with the tree.  For example n = 3374961778893 is a
\\ caterpillar and needs at most prime(8455786) which is no trouble.
\\
\\ On the other hand, a relatively small n = 408 has more vertices and asks
\\ for prime(252079129) which is over 5 billion and and beyond what a 32-bit
\\ PARI build can easily calculate (circa PARI 2.13).
\\
\\ There's no attempt to watch for repeats of subtrees numbers since repeats
\\ of large subtrees are rare in a "random" tree and getting some fast
\\ prime(i) will be better value.

\\-----------------------------------------------------------------------------
print("end");
