\\ Copyright 2019 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0);
default(strictargs,1);
read("../vpar.gp");


\\ Usage: gp <footprints-restricted.gp
\\
\\ This is a variation on footprint-preorder.gp (see references there)
\\ considering Knuth's original answer, before errata of 23 Apr 19.
\\
\\ The pre-errata answer was too small.  It counted over only some of the
\\ preorder forests, not all of them.  The code below is a couple of
\\ approaches to classes of trees or forests that answer did count over.


\\-----------------------------------------------------------------------------
\\ These helpers are the same as in footprint-preorder.gp.
\\ See comments in that file.

\\ footprint is a vector of 1,0 flags.
\\ Return [pvec,qvec] vector of two vectors pvec and qvec.
footprint_to_pqvecs(footprint) =
{
  my(plist=List([]),
     qlist=List([]),
     i=1);
  while(i<=#footprint,
        my(p=0); while(i<=#footprint && footprint[i]==1, p++;i++);
        my(q=0); while(i<=#footprint && footprint[i]==0, q++;i++);
        listput(plist,p);
        listput(qlist,q));
  [Vec(plist), Vec(qlist)];
}

\\ b is an integer.  Return a footprint vector of 0s and 1s.
b_to_footprint(b) =
{
  b >= 1 || error();
  concat(binary(b)[^1], [0]);
}

\\ vpar is a preorder labelled forest.  Return its childful footprint.
\\ This is equivalent to vector(#vpar,v, vpar_is_childful(vpar,v)),
\\ but here calculated from all num_children at once.
\\ (Perhaps this could be a vpar_is_childful_vector() func in the future.)
\\
footprint_of_vpar(vpar) = \
  apply(c->c!=0, vpar_num_children_vector(vpar));


\\------------------------------------------------------------------------------
\\ Knuth's pre-errata answer to exercise 30(b) counting forests of given
\\ footprint is count_restricted_footprint().

\\ Catalan numbers (A000108)
Catalan_number(n) = binomial(2*n,n) / (n+1);

\\ 7.2.1.6 exercise 30(b) answer first edition page 788,
\\ or draft fasc4a.ps.gz page 53
count_restricted_footprint(footprint) =
{
  \\ footprint_to_pqvecs() is full lengths of 1,0 runs.
  my(pvec,qvec); [pvec,qvec]=footprint_to_pqvecs(footprint);
  my(k = #pvec);
  for(i=2,k, pvec[i]--);  \\ reduce all -1 except pvec[1]
  for(i=1,k, qvec[i]--);
  prod(i=1,k, binomial(pvec[i]+qvec[i], qvec[i])) * Catalan_number(k-1);
}

{
  \\ Exercise 30(b) question footprint and pre-errata answer 240240.
  my(footprint
   =[1,0, 1,0, 1,1,0, 1,1,1,1,1,1,0,0,0,0, 1,0, 1,0, 1,0,0,0, 1,0, 1,1,0,0,0]);
  my(got=count_restricted_footprint(footprint));
  print("Knuth 30(b) pre-errata answer "got);
  got==240240 || error();
}

{
  my(lo=8,hi=31);
  print("Count restricted for b="lo" to "hi);
  for(b=lo,hi,
     my(footprint=b_to_footprint(b));
     print1(count_restricted_footprint(footprint),","));
  print("\n");

  \\ not in OEIS: 1,1,1,1,3,1,3,1,1,1,1,1,1,2,2,1,4,2,1,1,6,1,4,1,
}

{
  \\ The first footprint where restricted formula differs from all preorders
  \\ is n=4 vertices
  \\
  \\     footprint = [0, 1, 0, 0]
  \\
  \\ The restricted formula is, with -1s applied, p[1]=0, q[1]=0,
  \\ p[2]=0, q[2]=1.
  \\
  \\     binomial(0+0,0) * binomial(0+1,1) * Catalan_number(1) = 1
  \\
  \\ But there are 2 preorder forests with this footprint, being 2 places to
  \\ put the last vertex 4.
  \\
  \\     1 2       1 2 4
  \\       |\        |
  \\       3 4       3
  \\
  \\ Also in n=4 vertices, footprint=[1,0,1,0] is count 1 vs 2 in a similar
  \\ way.  It is, with -1s applied, p[1]=1, q[1]=0, p[2]=0, q[2]=0 for same
  \\ binomials (swapped).  The forests of this footprint are
  \\
  \\     1 3       1
  \\     | |       |\
  \\     2 4       2 3
  \\                 |
  \\                 4

  \\ my(footprint);
  \\ for(b=2,2^4-1,
  \\    footprint=b_to_footprint(b);
  \\    if(count_footprint_using_upwards(footprint)
  \\       != count_restricted_footprint(footprint), break));
  \\ print("Knuth formula first different on footprint = "footprint);
  \\ print("  formula "count_restricted_footprint(footprint));
  \\ print("  believe "count_footprint_using_upwards(footprint));
  \\ print();
}


\\-----------------------------------------------------------------------------
\\ Restricted Total
\\
\\ The way the restricted formula is too small can be seen in its grand
\\ total over all footprints of length n.
\\
\\ Footprints must have last entry 0, and other entries any 0,1 for 2^(n-1)
\\ combinations.  Total count trees of all these footprints is OEIS A086622.

A086622_samples = \
  [1, 2, 5, 12, 30, 76, 197, 520, 1398, 3820, 10594, 29768, 84620, 243000];
{
  print("Total restricted length n, OEIS A086622");
  for(n=1,10,
     my(total=0);
     forvec(footprint=vector(n,i, [0,i!=n]),
           total += count_restricted_footprint(footprint));
     print1(total,",");
     total == A086622_samples[n] || error());
  print("\n");
}

\\ Jose Luis Ramirez Ramirez in OEIS A086622 gives the following formula for
\\ A086622, adapted here to sum with index k starting k=1.
total_restricted(n) = \
  sum(k=1,floor((n+1)/2), 2^(n+1-2*k)*binomial(n-k,k-1)*Catalan_number(k-1));
{
  for(n=1,#A086622_samples,
     A086622_samples[n] == total_restricted(n) || error());
}







  \\ my(footprint);
  \\ for(b=2,2^4-1,
  \\    footprint=b_to_footprint(b);
  \\    if(count_footprint_using_upwards(footprint)
  \\       != count_restricted_footprint(footprint), break));
  \\ print("Knuth formula first different on footprint = "footprint);
  \\ print("  formula "count_restricted_footprint(footprint));
  \\ print("  believe "count_footprint_using_upwards(footprint));
  \\ print();


\\-----------------------------------------------------------------------------
\\ Total Number of Tree of Chains Starting 1
\\
\\ Some trees of chains have footprints starting with 1, meaning its first,
\\ as a tree, is p[1] >= 2.  Looks like this is cumulative A086622.

cumulative_A086622_samples = \
  vector(#A086622_samples,n,sum(i=1,n,A086622_samples[i]));
\\ not in OEIS: 1, 3, 8, 20, 50, 126, 323, 843, 2241, 6061

{
  print("Total restricted, footprint start with 1, is cumulative 1 to n-1");
  for(n=1,10,
     my(total=0);
     forvec(footprint=vector(n,i, if(i==1,[1,1], i==n,[0,0], [0,1])),
           total += count_restricted_footprint(footprint));
     print1(total,",");
     if(n>=2,
        total == cumulative_A086622_samples[n-1] || error());
  );
  print();
}


quit

\\-----------------------------------------------------------------------------
\\ Tree of Chains
\\
\\ One way to cook up a forest form of n vertices to match the formula is to
\\ restrict as follows.  It's convenient to express in terms of a "tree of
\\ chains" with n+1 vertices, then drop the tree root R for forest n.
\\
\\      0 top=R
\\          |\   \       \
\\      1   * Q   top     top         tree of chains
\\          |\     |     \ or
\\      2   * Q    *      top
\\          |\
\\      3   *
\\   p=4    |\
\\          L Q  q=4
\\
\\ Childless vertices Q can go under any of their own p run.  If the
\\ previous run is higher up then it and possible other preceding would be
\\ open to an unrestricted pre-order, but for tree of chains that are not
\\ allowed, must Q under own run of p childful vertices.
\\
\\ Each top vertex (after the root) of a run is restricted to be a child of
\\ another top.  This has the effect that the tops, with everything else
\\ ignored, are a tree.
\\
\\ The root of this n+1 tree is always childful.  Discarding that root gives
\\ a forest, and the footprint is based on that.
\\
\\ It can be noticed the restrictions, if expressed in terms of the forest,
\\ would be a little odd.  The Qs of the first run are allowed to be new
\\ roots, but not later Qs.  Top must be child of top, except the first
\\ forest vertex (the second tree vertex) does not count as a top for that
\\ purpose.  Better to think in terms of tree n+1.
\\
\\ The arrangements of tops as tree is vpar_count_ordered(n) = Catalan(n-1)
\\ ways.  The arrangement of Qs under their respective runs are q-1 things
\\ into p places, which is binomial(p-1+q-1,q-1) as per Knuth's p+q,q with
\\ Knuth's p,q each -1 except the first p[1].
\\
\\ The example footprint [0,1,0,0] above would be as follows, and only the
\\ first forest is tree of chains form.  The second is disallowed because
\\ the childless 4 must go somewhere under its own chain, which is just
\\ vertex 2.  4 cannot go higher as child of R.
\\
\\     R               R
\\     |\  good        |\ \  not tree of chains
\\     1 2             1 2 4
\\       |\              |
\\       3 4             3
\\
\\ The second example footprint [1,0,1,0] above is also second forest not
\\ tree of chains, since its top 3 is child of non-top 1 (can only be child
\\ of top R).


\\ vpar is a preorder labelled tree.
\\ Return 1 if it has the "tree of chains" form described above.
is_treeofchains(vpar) =
{
  my(num_children=vpar_num_children_vector(vpar),
     is_top=vector(#vpar),
     prev_top=1);
  is_top[1]=1;
  for(v=2,#vpar,
     if(num_children[v]==0,
        vpar[v] >= prev_top || return(0)); \\ childless must be under own run
     if(num_children[v]!=0 && num_children[v-1]==0,
        \\ new top
        is_top[v]=1;
        is_top[vpar[v]] || return(0);   \\ top must have parent also top
        prev_top=v));
  1;
}

\\ vpar is a preorder labelled tree, expected to be "tree of chains" form.
\\ Return its childful footprint for it as a forest, meaning discard its root.
\\
footprint_of_vpar_tree(vpar) = footprint_of_vpar(vpar)[^1];

{
  print("Tree-of-chains counts by footprint, per Knuth's formula");
  for(n=3,5,
     print("n="n);
     my(total=0);
     forvec(footprint=vector(n,i, [0,i!=n]),
           my(count=count_restricted_footprint(footprint));
           print("  "footprint"  "count);
           total += count);
     print("  total "total));
}
{
  print1("checking ... ");
  for(n=2,10,
     my(footprints=List([]));    \\ all possible footprints of n vertices
     forvec(footprint=vector(n,i, [0,i!=n]),
           listput(footprints,footprint));
     footprints=Set(footprints);
     my(counts=vector(#footprints));
     #footprints == 2^(n-1) || error();

     my(vpar=vpar_preorder_first(n+1));
     until((vpar=vpar_preorder_next(vpar))=='none, \\ trees
           \\ print(vpar" "footprint_of_vpar_tree(vpar)"  "is_treeofchains(vpar));
           is_treeofchains(vpar) || next;             \\ of necessary type
           my(footprint=footprint_of_vpar_tree(vpar),
              pos=setsearch(footprints,footprint));
           pos || error("oops, footprint not found: "footprint);
           counts[pos]++);

     for(i=1,#footprints,
        counts[i]!=0 || error();
        count_restricted_footprint(footprints[i]) == counts[i] || error());
  );
  print("good");
  print();
}


\\-----------------------------------------------------------------------------