\\ Copyright 2019, 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0);
default(strictargs,1);
read("../vpar.gp");


\\ Usage: gp <Huffman-1-to-n.gp
\\
\\ A binary Huffman code is a set of symbols which are represented by
\\ strings of 1s and 0s of various lengths.  No string is a prefix of
\\ another ("prefix-free") so that a symbol can be recognised immediately
\\ when its bits are seen, as nothing else starts the same.  For example
\\
\\     A = 00
\\     B = 01
\\     C = 1
\\
\\ These strings are descents down a binary tree.  Each vertex has two
\\ children, an 0 and a 1.  The symbols are the leaf "external" vertices.
\\
\\          *
\\        0/ \1
\\        *   C
\\      0/ \1
\\      A   B
\\
\\ Symbols can have strings chosen so that frequently occurring symbols have
\\ short strings, and infrequently occurring symbols have long strings.
\\ Frequency of occurrence is a "weight" at a leaf of the tree, and the
\\ weight at an internal is the total of the leaves below it.
\\
\\ The "greedy" algorithm for forming strings from symbol frequencies is to
\\ remove the two smallest weights and distinguish them by 0 or 1 under a
\\ new vertex.  The weight of the new vertex is the sum of two weights
\\ taken.  This new weight goes back into the set of weights for subsequent
\\ consideration of the next two smallest.
\\
\\ The codes formed here are for symbols numbered 1 to n and with those
\\ symbols having frequencies 1 to n respectively.  For example
\\
\\           [15]
\\          /    \           symbols v = 1 to 5
\\        [6]     [9]        weights     1 to 5
\\        / \     / \        merged total weight below is shown [3] etc
\\       3  [3]  4   5
\\          / \
\\         1   2
\\
\\ The algorithm starts with weights 1..5.  The two smallest are 1,2, so sum
\\ them as new weight 3.  The two smallest are now 3,3 sum them to 6.  Then
\\ 4,5 sum to 9.  Finally 6,9 sum to 15.
\\
\\ When taking two smallest weights, there's a free choice which is to be
\\ the 0-bit and which the 1-bit.  In the program code below, the
\\ distinction is made by the vertex numbers used.  The leaves are 1..n and
\\ the internals are n+1..2n-1 and the smaller vertex number is the 0-bit.
\\
\\ If a vertex was formed as a left-child right-child pair (like a Lisp
\\ cons-cell) then that would distinguish 0 and 1 descent and could
\\ construct any bit combinations.  The same could be done in the vpar by
\\ some relabelling if desired, but usually it's lengths of codes which are
\\ of interest rather than the particular bits.
\\
\\ Tree Uniqueness
\\ ---------------
\\
\\ If the 2nd and 3rd smallest weights are the same then there's a choice of
\\ which one to combine with the smallest.  Or similarly 2 or more equal
\\ second-smallest (or equal smallest if 1st and 2nd smallest are equal
\\ too).  The new weight is the same for any choice, but the tree structure
\\ under each vertex may differ so might result in a different tree.
\\
\\        v1      v2       v3          vN
\\     smallest   2nd  ==  3rd  ...  biggest          weight
\\       
\\ But starting from 1..n, weights 2nd==3rd do not occur.  Initial smallest
\\ weights < n go in a pattern
\\
\\     1, 2    --> 3
\\     3, [3]  --> 6               [3] = weight from 1+2
\\     4, 5    --> 9                    
\\     6, [6]  --> 12
\\     7, 8    --> 15
\\     9, [9]  --> 18
\\    10, 11   --> 21
\\     ...
\\
\\ The new weights are successive multiples of 3, occurring either by two
\\ original terms 3k+1 + 3k+2, or original 3k + new weight 3k.
\\
\\ The new weights are strictly increasing.  New weights are certainly never
\\ decreasing, since they sum non-decreasing terms.  Then they are strictly
\\ increasing unless there are four 1st=2nd=3rd=4th to make two new weights
\\ equal, and that does not occur within the 1..n pattern, and after that
\\ weights are strictly increasing.


\\-----------------------------------------------------------------------------

\\ Return a binary Huffman tree of n symbols with weights 1..n.
\\ Vertices 1..n are childless and are the "externals" of the tree.
\\ Vertices n..2n-1 are "internals" giving code structure.
\\ Vertex 2n-1 is the root.
\\ The internals have 2 children each (except n=1 has childless root).
\\
make_Huffman_n(n) =
{
  \\ v_list[i] = vertex number.
  \\ w_list[i] = weight of vertex number v_list[i], kept sorted in descending
  \\    order so the smallest weight is at the end.
  \\ Initially v_list is vertex numbers n..1 and w_list is their weights
  \\ n..1.  Then weights are combined to make new vertices.
  \\
  \\ p = last vertex number used, so vertex p+1 is available for use next.
  \\   Initially p=n.
  \\
  \\ The loop removes the two smallest weights.  They are made children of
  \\ new vertex p+1, and p+1 has weight the sum of its child weights.
  \\
  \\ p+1 is inserted into w_list[],v_list[] at the appropriate place to keep
  \\ them sorted in descending order of weight.  As described under "Tree
  \\ Uniqueness" above, the only equal weights which occur will be taken
  \\ together as two smallest, so it doesn't matter which way around they
  \\ are.  The code inserts on the larger side of any equal(s).
  \\
  \\ For greater efficiency could use a linked-list instead of listinsert(),
  \\ and could remember where the last insert was, since each w is
  \\ increasing.

  my(vpar=vector(2*n-1),
     w_list=List(Vecrev([1..n])),
     v_list=w_list,
     p=n,
     k);
  while((k=#w_list)>=2,
        \\ print(Vec(w_list)" "Vec(v_list));

        for(i=1,#w_list-1,
           lex([w_list[i],v_list[i]],[w_list[i+1],v_list[i+1]]) > 0
             || error("oops, w_list is meant to be descending order"));
        
        p++;
        vpar[v_list[k]] = vpar[v_list[k-1]] = p;  \\ children of new p
        my(w = w_list[k] + w_list[k-1]);          \\ new weight at p
        listpop(w_list); listpop(w_list);
        listpop(v_list); listpop(v_list);
        forstep(pos=#w_list,1,-1,              \\ from smallest to largest
           if(w_list[pos]>w,
              \\ print("  popped to "Vec(w_list)" "Vec(v_list));
              \\ print("  insert at pos="pos" cf "w_list[pos]);
              listinsert(w_list,w,pos+1);  \\ immediately after pos
              listinsert(v_list,p,pos+1);
              next(2)));
        listinsert(w_list,w,1);   \\ w is the biggest weight
        listinsert(v_list,p,1));
  vpar;
}

\\ The code for a symbol s=1..n is found by going down from the root to s
\\ and writing 0 when to the left (smaller) child or 1 when to the right
\\ (bigger) child, at each internal vertex.
\\
\\ code_of_symbol() return a vector of bits which are the code of symbol s.
\\ This is implemented using vpar_siblingpos_of_vertex() which is 0 or 1 for
\\ smaller or bigger child position.
\\
\\ For efficiency, might want to get the whole vpar_siblingpos_vector() at
\\ once, instead of individual vpar_siblingpos_of_vertex().  That would be
\\ so especially if making the codes for several symbols.  But individual
\\ vpar_siblingpos_of_vertex() is convenient.
\\
code_of_symbol(vpar,s) =
{
  my(l=List([]));
  while(vpar[s],
        listput(l,vpar_siblingpos_of_vertex(vpar,s));
        s=vpar[s]);
  Vecrev(l);
}

{
  print("Huffman trees vpar vectors");
  for(n=1,7,
     my(vpar=make_Huffman_n(n));
     print("  n="n"  "vpar);
     if(n>=2,
        forstep(s=n,1,-1,
           my(vec=code_of_symbol(vpar,s));
           print("    s="s" code ",concat(apply(bit->Str(bit),vec))))));
  print();
}

{
  for(n=1,50,
     my(vpar=make_Huffman_n(n));
     vpar_validate(vpar);
     vpar_num_childless(vpar)==n || error();
     );
}

\\-----------------------------------------------------------------------------
{
  print("Tree height = maximum codeword length");
  print1("  ");
  for(n=3,32,
     my(vpar=make_Huffman_n(n),
        height=vpar_height(vpar));
     print1(height, ","));
  print("   (A126236)");
}
{
  my(A126236_samples=[1,2,3,3,4,4,5,5,5,5,6,6,6,6,7,7,7,7,7,7,7,7,
                      8,8,8,8,8,8,8,8,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,
                      10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,
                      11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,
                      11,11,11,11,11]);
  vector(#A126236_samples,n,n++; vpar_height(make_Huffman_n(n)))
    == A126236_samples  || error();
}

{
  \\ Minimum depth of a childless vertex is the minimum code length.
  \\ In vpar.gp this is vpar_protnum().
  \\
  print("Tree protnum (min depth childless) = minimum codeword length");
  print1("  ");
  for(n=3,32,
     my(vpar=make_Huffman_n(n));
     print1(vpar_protnum(vpar), ","));
  print("   (A126235)");
  print();
}
{
  my(A126235_samples=[1,1,1,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,
                      4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
                      5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
                      5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
                      6,6,6,6,6,6,6,6,6]);
  vector(#A126235_samples,n,n++; vpar_protnum(make_Huffman_n(n))) ==
          A126235_samples  || error();
}

\\-----------------------------------------------------------------------------
{
  print("Depth of v=1..n is its codeword length    (A126268, for n>=3)");
  for(n=3,8,
     my(vpar=make_Huffman_n(n),
        depths=vpar_depths_vector(vpar));
     print1("  n="n":  ");
     for(v=1,n,
        print1(depths[v],","));
     print());
  print("");
}
{
  my(A126268_samples=[2,2,1,3,3,2,1,3,3,2,2,2,4,4,3,2,2,2,4,4,3,3,3,2,2,5,5,
                      4,3,3,3,2,2]);
  my(got=[]);
  for(n=3,#A126268_samples,
     got=concat(got, vpar_depths_vector(make_Huffman_n(n))[1..n]);
     if(#got>=#A126268_samples, break));
  Vec(got,#A126268_samples) == A126268_samples  || error();
}

\\-----------------------------------------------------------------------------
{
  print("List of symbols where codeword length grows are A126014, and count A126237.");
  print("This is v = n..1 where depth increases.");
  for(n=3,8,
     my(vpar=make_Huffman_n(n),
        depths=vpar_depths_vector(vpar),
        count=0);
     print1("  n="n":  ");
     forstep(v=n-1,1,-1,
        if(depths[v] > depths[v+1],  \\ increase in codeword length
           print1(v,",");
           count++));
     printf("%*s count %d\n", 10-2*count, "", count));
  print("");
}
{
  my(A126014_samples=[2,3,2,2,3,2,5,2,6,3,2,8,3,2,9,5,2,5,2,6,3,2,8,3,2,9,
                      5,2,11,5,2,12,6,3,2,14,6,3,2,15,8,3,2,17,8,3,2,18,9,
                      5,2,20,9,5,2,21,11,5,2,11,5,2,12,6,3,2,14,6,3,2,15,8,
                      3,2,17,8,3,2,18,9,5,2,20,9,5,2,21,11,5,2]);
  my(l=List([]));
  for(n=3,#A126014_samples,
     my(vpar=make_Huffman_n(n),
        depths=vpar_depths_vector(vpar));
     forstep(v=n-1,1,-1,
        if(depths[v] > depths[v+1],  \\ increase in codeword length
           listput(l,v);
           if(#l == #A126014_samples, break(2)))));
  Vec(l) == A126014_samples  || error();

  my(A126237_samples=[1,2,1,2,2,3,3,3,2,3,3,3,3,4,4,4,4,4,4,4,3,
                      4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,4,
                      5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
                      6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                      6,6,6,6,5,6,6,6,6,6,6,6,6,6]);
  vector(#A126237_samples,n,n+=2;
         my(vpar=make_Huffman_n(n),
            depths=vpar_depths_vector(vpar));
         sum(v=1,n-1, depths[v] > depths[v+1]))
    == A126237_samples  || error();
}

\\-----------------------------------------------------------------------------
print("end");
