\\ Copyright 2020, 2021 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0);
default(strictargs,1);
read("../vpar.gp");


\\ Usage: gp <complete-binary-matchings.gp
\\
\\ This is a mixture of code, formulas, descriptions, and workings-out, for
\\ the number of matchings in a complete binary tree.
\\
\\ The resulting growth power is related to limits in Wolfram's 1-D cellular
\\ automaton rule 150 and which is odd terms in trinomial powers (1+x+x^2)^n.


\\-----------------------------------------------------------------------------
\\ Generic Things

nearly_equal(x,y, epsilon=1e-8) = abs(x-y) < epsilon;
to_binary(n) = fromdigits(binary(n),10);
from_binary(n) = fromdigits(digits(n),2);
bitlength(n) = if(n,logint(n,2)+1,0);
log2(n) = log(n)/log(2);


\\-----------------------------------------------------------------------------
\\ Matchings
\\
\\ A complete binary tree is rows like
\\
\\          *         row 1  (root)
\\        /   \
\\       *     *      row 2
\\      / \   / \
\\     *   * *   *    row 3
\\
\\ A matching is a set of vertex pairs where the two vertices of a pair are
\\ connected by an edge, and a vertex is used in at most one pair.  There
\\ are, in general, many ways to select such pairs.  Any number of pairs can
\\ be assembled, including no pairs (the empty matching).
\\
\\ A recurrence for the number of matchings in the complete binary tree of n
\\ rows follows by considering the root either matched or unmatched, and
\\ below it subtrees of smaller sizes n-1 or n-2 rows.
\\
\\     num rows
\\        n       unmatched         matched       matched
\\                  /   \             //  \         / \\
\\       n-1      any   any      matched  any     any  matched
\\                                  / \                / \
\\       n-2                      any any            any any
\\
\\         a(n) =  a(n-1)^2   +     2 * a(n-1) * a(n-2)^2
\\
\\ This recurrence starts from a(0) = a(1) = 1 which is just 1 matching, the
\\ empty matching, in a tree of no rows or 1 row.

a(n) =
{
  n>=0 || error("a() is for n>=0");
  my(x=1,y=1);
  for(i=2,n,
     [x,y] = [x^2 + 2*x*y^2, x]);
  x;
}
{
  print("Complete binary tree number of matchings (n rows)");
  print1("  a(n) = ");
  for(n=0,6, print1(a(n),", "));
  print("...   n>=0");
}
{
  a(0) == 1     || error();   \\ empty tree, no rows, empty matching only
  a(1) == 1     || error();   \\ singleton,  one row, empty matching only
  a(2) == 3     || error();   \\ path-3, first pair, second pair, or empty
  for(n=2,6,
     a(n) == a(n-1)^2 + 2*a(n-1)*a(n-2)^2  || error());
}

\\ vpar_make_complete_tree() creates a complete tree of given height,
\\ which is height = num_rows - 1.
\\
\\ There's alway a bit of a choice whether n rows or height n.  An
\\ attraction of n rows is that 0 rows is 0 vertices (and 1 rows is 1
\\ vertex).  On the other hand, height n is the vpar_height() property.
\\ Here use 0 rows 0 vertices for thinking about matchings.
\\
make_complete_binary_rows(n) = vpar_make_complete_tree(2,n-1);
{
  make_complete_binary_rows(0) == []       || error();
  make_complete_binary_rows(1) == [0]      || error();
  make_complete_binary_rows(2) == [0,1,1]  || error();
}

\\ vpar_matchings_count() counts the matchings in a vpar
{
  for(n=0,10,
     my(vpar=make_complete_binary_rows(n));
     a(n) == vpar_matchings_count(vpar)    || error());
}

print();


\\-----------------------------------------------------------------------------
\\ Products of Jacobsthal Numbers

print("Products of Jacobsthals");

\\ The Jacobsthal numbers are
J(n) =
{
  n>=0 || error("J() is for n>=0");
  (2^n - (-1)^n)/3;
}
{
  print1("  J(n) = ");
  for(n=1,8, print1(J(n),", "));
  print("...   n>=1  (A001045)");
}
{
  \\ which is OEIS A001045
  my(A001045_samples=[0,1,1,3,5,11,21,43,85,171,341,683,1365,2731,5461]);
  vector(#A001045_samples,n,n--; J(n)) == A001045_samples
    || error();   \\ A001045_samples OFFSET=0
}

\\ a(n) can be written as a product,
\\
\\     a(n) = J(n+1) * J(n) * J(n-1)^2 * J(n-1)^4 * ... * J(1)^(2^(n-1))
\\
\\                      n-1
\\          = J(n+1) * prod  J(n-i)^(2^i)
\\                      i=0
\\
\\ The pattern is initial J(n+1) and then J(n) downwards with successively
\\ doubling powers starting from 1.  Since J(1) = J(2) = 1, the product can
\\ stop at J(3) if desired.  But doing so would make initial n=0 and n=1
\\ cases look like exceptions to the general rule.
{
  a(0) == J(1)                                                || error();
  a(1) == J(2) * J(1)                                         || error();
  a(2) == J(3) * J(2) * J(1)^2                                || error();
  a(3) == J(4) * J(3) * J(2)^2 * J(1)^4                       || error();
  a(4) == J(5) * J(4) * J(3)^2 * J(2)^4 * J(1)^8              || error();
  a(5) == J(6) * J(5) * J(4)^2 * J(3)^4 * J(2)^8 * J(1)^16    || error();

  for(n=0,12,
     a(n) == J(n+1) * prod(i=0,n-1, J(n-i)^(2^i))   \\ terms n-i = n downwards
       || error());
}
{
  for(n=2,5,
     my(p = J(n+1)*J(n));
     printf("  a(%d) = %6d = %d * %d", n, a(n), J(n+1), J(n));
     for(i=1,n-1,
        my(j=J(n-i), e=2^i); p *= j^e;
        print1(" * ",j,"^",e));
     print();
     p == a(n) || error());
}

\\ This product form follows by induction from the recurrence for a(n).
\\ Suppose the product form is true up to n-1, then the products for a(n-1)
\\ and a(n-2) in the recurrence for a(n) are
\\
{
  for(n=2,12,
     \\ "unmatched" part
     a(n-1)^2 == J(n)^2*J(n-1)^2 * prod(i=2,n-1, J(n-i)^(2^i))   || error();

     \\ "matched" part
     a(n-2)^2 == J(n-1)^2        * prod(i=2,n-1, J(n-i)^(2^i/2)) || error();
     a(n-1)   == J(n) *J(n-1)    * prod(i=2,n-1, J(n-i)^(2^i/2)) || error();
     2*a(n-1)*a(n-2)^2 == 2 * J(n) * J(n-1)^3
                                 * prod(i=2,n-1, J(n-i)^(2^i)) || error();

     \\ total
     a(n)     == ( J(n)^2*J(n-1)^2 + 2 * J(n) * J(n-1)^3 )
                                 * prod(i=2,n-1, J(n-i)^(2^i)) || error();
     a(n)     == ( J(n) + 2*J(n-1) )
                                 * prod(i=0,n-1, J(n-i)^(2^i)) || error();
     a(n)     ==   J(n+1)        * prod(i=0,n-1, J(n-i)^(2^i)) || error();

     \\ final J(n+1) = J(n) + 2*J(n-1) is a usual Jacobsthal recurrence,
     \\ or follows from its powers
     J(n+1) == J(n) + 2*J(n-1)  || error();
  );
}

\\ A simple consequence is that a(n) is a divisor of any bigger a(N).  So a
\\ "divisibility sequence", meaning a(n) divides a(n*m) for any n and m.
\\
for(big=1,8, for(small=1,big-1, a(big) % a(small) == 0  || error()));

\\ Clark Kimberling defines a "strong divisibility sequence" as a sequence
\\ having gcd(a(m),a(n)) = a(gcd(m,n)).  Maybe a(n) here is such a sequence.
\\ Certainly a(gcd(m,n)) is a divisor of both a(m) and a(n), for the simple
\\ reason that gcd(m,n) <= min(m,n).  But is a(gcd(m,n)) the greatest common
\\ divisor?  Could a(m) and a(n) have more factors in common than just
\\ a(gcd(m,n)) ?

\\ strong divisibility holds for small n,m at least ...
for(n=1,10, for(m=1,n, gcd(a(n),a(m)) == a(gcd(n,m) || error())));

print();


\\-----------------------------------------------------------------------------
\\ Matchings Asymptotic
\\
\\ The asymptotic growth of a(n) is
\\
\\     a(n) -> 1/2 * C^(2^n)
\\
\\ where growth power
\\
\\     C =   limit   c(n)  = 1.537176717...     (OEIS A338294)
\\         n->infinity
\\
\\     with c(n) = (2*a(n))^(1/2^n)

c(n) = (2*a(n))^(1/2^n);
C = c(20);  \\ an approximation

{
  \\ eg. n=6  445957080991.18 vs 448046589375  ratio 0.99534
  printf("Asymptotic a(n) -> 1/2*C^(2^n) where C = approx %.7f...\n", C);
  for(n=0,6,
     my(exact=a(n),
        asymp=1/2*C^(2^n));
     printf("  n=%d %16.2f vs %-12d  ratio %.5f\n",
            n, asymp, exact, asymp/exact));
  print();
}

\\ Fixed factor 1/2 is the easier part.  It has no effect on the n'th root
\\ for the growth power, in the sense that say a(n)^(1/2^n) -> C too.
\\ Any constant F has F^(1/2^n) -> 1 as n->infinity.  Factor 1/2 brings the
\\ asymptotic closer to the real result sooner for small n (the printout
\\ above).  Factor 1/2 is since
\\
\\                     a(n)
\\       limit       --------   = 1/(1/2)  = 2
\\     n->infinity   a(n-1)^2
\\
\\ To see this, write
f(n) = a(n)/a(n-1)^2;

{
  print("Fixed factor limit approached by ratios of Jacobsthals");
  print1("  f(n) = a(n) / a(n-1)^2\n       = ");
  for(n=1,7, print1(a(n) / a(n-1)^2,", "));
  print("...   n>=1");
}

\\ Then, using the Jacobsthal products form of a(n), terms cancel in f(n)
\\ leaving a ratio of Jacobsthals, which approaches 2.
{
  for(n=2,10,
     f(n) == J(n+1) / J(n)     || error();
     f(n) == 2  + (-1)^n/J(n)  || error());
}

\\ The recurrence for a(n), divided through by a(n-1)^2, gives a recurrence
\\ for f(n).
\\
\\     a(n)/a(n-1)^2 = (a(n-1)^2 + 2*a(n-1)*a(n-2)^2) / a(n-1)^2
\\                   =  1        + 2/ (a(n-1)/a(n-2)^2)
\\
\\     f(n)          =  1        + 2/ f(n-1)
\\
{
  for(n=2,10,
     a(n)/a(n-1)^2 == (a(n-1)^2 + 2*a(n-1)*a(n-2)^2) / a(n-1)^2     || error();
     a(n)/a(n-1)^2 ==  1        + 2/ (a(n-1)/a(n-2)^2)              || error();
     f(n)   == 1 + 2/f(n-1)                                         || error();
     f(n)/2 == 1/2 + 1/f(n-1)                                       || error();
     );
  nearly_equal(f(12),   2, 1e-3) || error();
  nearly_equal(f(12)/2, 1, 1e-3) || error();
}

\\ With this recurrence for f, the limit for f(n) would be an infinite
\\ continued fraction
\\
\\     limit f(n)  = 1 + 2/(1 + 2/(1 + 2/( ... )))           = 2
\\     n->inf
\\                 = [1, 1/2, 1, 1/2, ...] ordinary continued fraction
\\
\\ Any initial value f(1) is mapped ever closer to 2 by the 1 + 2/f(n-1)
\\ process.  Two steps go closer to 2 from above
\\
\\     x - 2 > 1 + 2/(1 + 2/x) - 2
\\     (x-2)*(x+1) / (x+2) > 0     when x > 2
\\
\\ or from below
\\
\\     2 - x > 2 - (1 + 2/(1 + 2/x))
\\     - (x-2)*(x+1) / (x+2) > 0   when 0 <= x < 2
\\
\\ From-below is the present case, since f(1)=1.
\\
f(1) == 1   || error();

print();


\\-----------------------------------------------------------------------------
\\ Asymptotic Growth Power C
\\
\\ For growth power C, to see that c(n) converges, use the Jacobsthal
\\ products form again.  The 2^n'th root divides each power by 2^n so
\\
\\     c(n) =   J(1)  ^(1/2)
\\            * J(2)  ^(1/4)
\\            * J(3)  ^(1/8)
\\            * ...
\\            * J(n)  ^(1/2^n)
\\            * J(n+1)^(1/2^n)
\\            * 2^(1/2^n)
{
  for(n=0,6,
     nearly_equal(c(n),
                  (2* J(n+1) * prod(i=0,n-1, J(n-i)^(2^i)))^(1/2^n) )
                 || error();

     nearly_equal(c(n),
                  2^(1/2^n)
                  * J(n+1)^(1/2^n)
                  * prod(i=0,n-1, J(n-i)^(2^(i-n))) )
                 || error();

     nearly_equal(c(n),
                  2^(1/2^n)
                  * J(n+1)^(1/2^n)
                  * prod(i=1,n, J(i)^(1/2^i)) )
                 || error();
     );
}

\\ It's convenient to take a log so sum of logs of the product terms
\\
\\     log(c(n)) =   log(J(1)) / 2
\\                 + log(J(2)) / 4
\\                 + log(J(3)) / 8
\\                 + ...
\\                 + log(J(n)) / 2^n
\\                 + log(J(n+1)) / 2^n
\\                 + log(2) / 2^n
{
  for(n=0,6,
     nearly_equal(log(c(n)), log(2*a(n))/2^n)  || error();

     nearly_equal(log(c(n)),
                    sum(i=1,n, log(J(i)) / 2^i) )
                  + log(J(n+1)) / 2^n
                  + log(2) /2^n
                 || error();
     );
}

\\ A simple upper bound J(n) < 2^n shows that
\\
\\     log(J(n))/2^n  <  log(2^n)/2^n
\\                    =  log(2) * n/2^n
\\
\\ and a split of the numerator "n" into n copies of 1/2^n gives a usual
\\
\\     sum n/2^n  =  1/2 + 1/4 + 1/8 + ...
\\     n>=1              + 1/4 + 1/8 + ...
\\                             + 1/8 + ...
\\                =  1                         \
\\                       + 1/2                 | sum each row
\\                             + 1/4 + ...     /
\\                =  2
\\
\\ so log(C) has an upper bound and is a sum of positive terms, so
\\ converges.
\\
{
  for(n=1,10,
     1/3*2^(n-1) < J(n)  || error();
     J(n) < 2^n          || error());
  nearly_equal(suminf(n=1, n/2^n), 2)  || error();
}
{
  print("log(c(n)) sum terms");
  my(total=0);
  for(i=1,7,
     my(l=log(J(i)) / 2^i);
     total += l;
     printf("  i=%d  %.8f\n", i, l));
  printf("  total %.8f\n", total);

  my(n=16);
  printf("    c(%d)  = approx %.7f...\n", n, c(n));
  printf("log(c(%d)) = approx %.7f...    -> limit A242049\n", n, log(c(n)));
}


\\-----------------------------------------------------------------------------
\\ log(C) = OEIS Constant A242049

\\ log(C) is constant A242049 in the OEIS, so C = exp(A242049) = A338294.
\\ 
\\     http://oeis.org/A242049
\\
\\                     inf   log( (2^(k+2) - (-1)^k)/3 )
\\     A242049 = 1/4 * sum   ---------------------------
\\                     k=1              2^k
\\
\\ This is numerator log(J(k+2)) and exactly the log(C) sum above, ignoring
\\ 
\\     log(J(1)) = log(J(2)) = 0
\\     log(J(n+1)) / 2^n    -> 0     extra over the sum as such
\\     log(2)/2^n           -> 0     final
\\
\\ Factor 1/4 on the A242049 sum is only since the terms are expressed in
\\ 2^(k+2).  The first term, k=1, is
\\
\\           log(J(3))     log(J(3))
\\     1/4 * ---------   = ---------   first non-zero term like here
\\              2^1           2^3

A242049_sample = 0.42994743334245272011469703551;

{
  for(k=1,10, J(k+2) == (2^(k+2) - (-1)^k) /3    || error());
  nearly_equal(log(J(1)), 0)                     || error();
  nearly_equal(log(J(2)), 0)                     || error();
  nearly_equal(log(J(3)), log(3))                || error();

  nearly_equal(A242049_sample, log(c(16)), 1e-6) || error();
  nearly_equal(A242049_sample,
               1/4 * sum(k=1,40, log((2^(k+2) - (-1)^k)/3) / 2^k))
               || error();
  nearly_equal(A242049_sample,   \\ bring in 1/4, and change of variable
               sum(k=3,40, log((2^k - (-1)^k)/3) / 2^k))
               || error();
  nearly_equal(A242049_sample,     \\ and numerator is J
               sum(k=3,40, log(J(k)) / 2^k))
               || error();

  \\ The sum can extend back to k=1 since J(1) = J(2) = 1 so their logs
  \\ are 0
  nearly_equal(A242049_sample,
               sum(k=1,40, log(J(k)) / 2^k))
               || error();
}


\\ log(J(k)) ~ k, so terms in the sum decrease by a little less than 1 bit
\\ each.  Finch, Sebah and Bai (reference below) mention Wynn's epsilon
\\ method for accelerating convergence.
\\
\\ A simple small improvement to convergence can be made by taking out a
\\ factor 2^k/3 from each J(k) so as to reduce terms from log(big) down to
\\ log(1+small),
{
  for(k=2,20,
     J(k) == 2^k/3 * (1 - (-1)^k/2^k)   || error());

  nearly_equal(A242049_sample,
               sum(k=1,40, log(2^k/3 * (1 - (-1)^k/2^k)) / 2^k)
              ) || error();
  nearly_equal(A242049_sample,
               sum(k=1,40, k*log(2)               / 2^k
                           - log(3)               / 2^k
                           + log(1 - (-1)^k/2^k)  / 2^k)
              ) || error();

  \\ then sum(k>=1,  1/2^k) = 1  in the usual way
  \\ and  sum(k>=1,  k/2^k) = 2  like above at the J(n) upper bound
  nearly_equal(suminf(k=1, 1/2^k), 1)  || error();
  nearly_equal(suminf(k=1, k/2^k), 2)  || error();

  nearly_equal(A242049_sample,
               log(4/3) + sum(k=1,40, log(1 - 1/(-2)^k) /2^k))
               || error();

  nearly_equal(A242049_sample,
               log(1+1/3) + log(1+1/2)/2
                          + log(1-1/4)/4
                          + log(1+1/8)/8
                          + log(1-1/16)/16
                          + log(1+1/32)/32,
               1e-2)  \\ converging only about 1 bit per term
               || error();
}

\\ Each of these log(1+small) suits Mercator's series for log,
\\
\\     log(1 + x) = -x + 1/2*x^2 - 1/3*x^3 + 1/4*x^4 - ...
\\                  alternating signs
\\
\\     log(1 - x) = -x - 1/2*x^2 - 1/3*x^3 - 1/4*x^4 + ...
\\                = sum(t=1,inf, -(1/t) *x^t)
\\                  all negatives
\\
\\ The log(4/3) part is like this too.  (But there's other much faster
\\ convergence for such a log isn't there?)
\\
\\    log(4/3) = log(1 - 1/(-3))
\\             = -1/(-3)^1 - (1/2)/(-3)^2 - (1/3)/(-3)^3 - (1/4)/(-3)^4 - ...
\\
nearly_equal(sum(t=1,15, -(1/t)/(-3)^t), log(4/3))  || error();

\\ The following code is log(4/3) by keeping an exact sum of terms so-far,
\\ scaled by 10^len desired decimal digits.  Since the sum is terms of
\\ alternating sign and decreasing absolute value, once two terms are both
\\ within the target range (same integer part of the scaled total) then
\\ further terms will not leave that range.
\\
log43_digits_vector(len) =
{
  my(p=10^len,    \\ power 10^len / (-3)^i, so terms wihout 1/i
     s=0,         \\ sum so far, will approach  10^len * log(4/3)
     prev=-1,     \\ integer part of s
     i=1);
  while(1,
        p/=-3;
        s-=p/i;
        my(this=floor(s));
        if(this==prev,  \\ unchanged integer part, have enough for the digits
           return(digits(this)));
        prev=this;
        i++);
}
{
  \\ A083679 = log(4/3) = 0.287682...
  my(A083679_samples=[2,8,7,6,8,2,0,7,2,4,5,1,7,8,0,9,2,7,4,3,9,2,1,9,0,0,5],
     len=#A083679_samples);
  log43_digits_vector(len) == A083679_samples  || error();
}

print();


\\-----------------------------------------------------------------------------
\\ log(C) Rest by Columns

\\ After taking out log(4/3), the rest is further logs
\\
\\               inf  log(1 - 1/(-2)^k)
\\    logCrest = sum  -----------------
\\               k=1         2^k
\\             = log(1+1/2)/2 + log(1-1/4)/4 + log(1+1/8)/8 + ...
\\             = 0.14226536089...
\\
logCrest_term_k(k) = log(1 - 1/(-2)^k) / 2^k;
logCrest(limit) = sum(k=1,limit, logCrest_term_k(k));
nearly_equal(logCrest(20) + log(4/3),  log(C))  || error();

\\ These can be written out with Mercator's series for log(1-x) again
\\
\\              t=1           t=2             t=3    column number
\\              ---           ---             ---
\\     row      -1*          -1/2*           -1/3*   factor in each column
\\     k=1:   1/(-2)^1    1/((-2)^1)^2    1/((-2)^1)^3  ...  /2^1
\\     k=2:   1/(-2)^2    1/((-2)^2)^2    1/((-2)^2)^3  ...  /2^2
\\     k=3:   1/(-2)^3    1/((-2)^3)^2    1/((-2)^3)^3  ...  /2^3
\\     k=4:   1/(-2)^4    1/((-2)^4)^2    1/((-2)^4)^3  ...  /2^4
\\
\\               inf   1   inf     -(1/t)
\\    logCrest = sum  ---  sum   ---------
\\               k=1  2^k  t=1   ((-2)^k)^t
\\
{
  for(k=1,5,
     \\ sum horizontally row k, stopped arbitrarily at t <= 20
     \\ though actual row continues infinitely
     nearly_equal(1/2^k * sum(t=1,20, -(1/t) / ((-2)^k)^t),
                  logCrest_term_k(k))   || error());
}

\\ Column t can be summed by switching the order of summation,
\\
\\               inf          inf        1
\\    logCrest = sum -(1/t) * sum  ----------------
\\               t=1          k=1  2^k * (-2)^(k*t)
\\
\\               inf          inf      (-1)^k
\\             = sum -(1/t) * sum  --------------
\\               t=1          k=1  (-2)^(k*(t+1))
\\
\\ Then a geometric sum of 1/( -(-2)^(t+1) )^k for k>=1 gives
\\
\\               inf       1/t
\\    logCrest = sum  --------------
\\               t=1  (-2)^(t+1) + 1
\\
{
  nearly_equal(sum(t=1,25, -(1/t) * sum(k=1,25-t, 1/(2^k * (-2)^(k*t) ))),
               logCrest(20))     || error();

  nearly_equal(sum(t=1,25, -(1/t) * sum(k=1,25-t, (-1)^k / (-2)^(k*(t+1)) )),
               logCrest(20))     || error();

  nearly_equal(sum(t=1,25, -(1/t) * sum(k=1,25-t, 1 / ( -(-2)^(t+1) )^k )),
               logCrest(20))     || error();

  nearly_equal(sum(t=1,25, 1/t / ((-2)^(t+1) + 1) ),
               logCrest(20))     || error();
}

\\ The result is then
\\
\\                        inf        1/t
\\    log(C) = log(4/3) + sum  --------------
\\                        t=1  (-2)^(t+1) + 1
\\
\\ The terms in this sum are again alternating signs so that again two
\\ partial sums within a target range means the limit is certainly within
\\ that range.
\\
\\ Convergence is a little over 1 bit per term, which is not particularly
\\ fast, but the terms are rationals.

\\ The following code is log(4/3) and logCrest calculated together using
\\ their respective alternating signs sums.
\\
\\              inf   1    /         1             1    \
\\    log(C) =  sum  --- * |  --------------  - ------  |
\\              t=1   t    \  (-2)^(t+1) + 1    (-3)^t  /
\\
\\ log43_term decreases as 1/3^t whereas logCrest_term decreases as 1/2^t,
\\ so log43_term becomes much smaller quite quickly.  A fancier loop might
\\ take a triplet of logCrest terms every so often so as to keep similar
\\ magnitude with the 1/3^t, but still alternating signs (the sign is the
\\ same in each).
\\
logC_digits_vector(len) =
{
  my(scale=10^len,   \\ multipled through the whole sum
     prev=0,         \\ sum so far, floor
     s=0,            \\ sum so far, as a rational
     t=1);           \\ term t
  while(1,
        my(log43_term    = -1/(-3)^t,
           logCrest_term = 1/((-2)^(t+1) + 1) );
        sign(log43_term)    == -(-1)^t  || error();  \\ alternating sign
        sign(logCrest_term) == -(-1)^t  || error();  \\ same
        s += scale/t * (log43_term + logCrest_term);
        my(this=floor(s));
        if(this==prev, return(digits(this)));
        prev=this;
        t++);
}
{
  \\ A242049 = log(C) = 0.429947433...
  \\ sample values from A242049
  my(A242049_samples=[4,2,9,9,4,7,4,3,3,3,4,2,4,5,2,7,2,0,1,1,4,6,9,7,
                      0,3,5,5,1,9,9,2,2,3,2,3,3,2,4,0,6,5,0,1,1,5,8,9,
                      3,0,4,6,1,7,0,4,0,2,7,6,0,7,2,5,7,4,2,8,3,3,7,2],
     len=#A242049_samples);
  logC_digits_vector(len) == A242049_samples  || error();
}

print();


\\-----------------------------------------------------------------------------
\\ Rule 150 Cells, T

\\ Stephen Wolfram defines certain 1-dimensional cellular automata in which
\\ each cell in a row is ON or OFF and in the next generation becomes ON or
\\ OFF according to its own state and the state of its two immediate
\\ neighbours each side.  (All cells step simultaneously to their new state.)
\\
\\ 3 cells determine the new state.  Each can be ON or OFF so 8 combinations.
\\ An automaton "rule" is an 8-bit number saying for each combination
\\ whether the cell becomes ON or OFF in the next generation.  "Row n" of
\\ the automaton is the states of each cell in generation n, starting from a
\\ single ON cell in generation 0.  Rule "150" starts
\\
\\     http://mathworld.wolfram.com/Rule150.html
\\
\\            11      ** * ** ** * ** ** * **
\\            10       * * *   * * *   * * *       bits A071036
\\             9        ***     ***     ***
\\             8         *       *       *         rows
\\             7          ** ** *** ** **          decimal A038184
\\             6           * *   *   * *           binary  A118110
\\             5            *** *** ***
\\             4             *   *   *             * = ON
\\             3              ** * **              blank = OFF
\\             2               * * *
\\             1                ***
\\     row n = 0                 *
\\
\\ (See also my Perl Math::PlanePath::CellularRule.)
\\
\\ Wolfram in
\\
\\     Stephen Wolfram, "Statistical Mechanics of Cellular Automata",
\\     Reviews of Modern Physics, volume 55, July 1983, pages 601-644.
\\     http://dx.doi.org/10.1103/RevModPhys.55.601
\\
\\     And author's copy:
\\     http://www.stephenwolfram.com/publications/cellular-automata-complexity/pdfs/statistical-mechanics-cellular-automata.pdf
\\
\\ calculates (top of page 614, or author's copy page 21) the number of ON
\\ cells in row n,
\\
\\     T(n) = prod  Chi(i)                          (OEIS A071053)
\\            run length i for each maximal consecutive 1-bits in n
\\
\\ where Chi(n) = J(n+2), so Jacobsthal numbers starting at Chi(1) = 3.
\\ For example run lengths,
\\
\\     n = binary  111 00 11111 0 11 0  = 14838
\\                 \-/    \---/   \/
\\                  3       5      2    run lengths
\\
\\     T(14838) = Chi(3) * Chi(5) * Chi(2) = 2365

Chi(n) = J(n+2);

Chi(3) * Chi(5) * Chi(2) == 2365  || error();

\\ Neil Sloane, "On the Number of ON Cells in Cellular Automata", 2015,
\\ section 3, http://arxiv.org/abs/1503.01168
\\
\\ defines a general "run length transform" for such a product.  An original
\\ sequence like Chi is transformed to a new sequence T by defining
\\ T(n) = product of Chi terms according to the run lengths of 1-bits in n.

\\ "func" is a function called as func(runlen) for various integer runlen >= 1
run_length_transform_of_func(n,func) =
{
  my(ret=1);
  for(k=0,bitlength(n)-1,
     if(bittest(n,k),             \\ start of run of 1s
        my(start=k);
        while(bittest(n,k++),);   \\ step up through those 1s
        ret*=func(k-start)));     \\ len = k-start, multiply func(len)
  ret;
}

T(n) = run_length_transform_of_func(n,Chi);
{
  my(n = from_binary( 111 00 11111 0 11 0 ));
  n == 14838   || error();
  T(n) == Chi(3) * Chi(5) * Chi(2)  || error();
}

\\ T(n) is OEIS sequence A071053.  run_length_transform_of_func() on Chi is
\\ essentially the same as Joerg Arndt's code in A071053, but shrunk down.

\\ An equivalent definition for T(n) is, per Wolfram again (bottom of page
\\ 606, or author's copy page 13), the number of odd coefficients in a
\\ trinomial power (1+x+x^2)^n
\\
for(n=0,256, T(n) == vecsum(Vec( (1+x+x^2)^n )%2)            || error());

\\ Looking for odd versus even coefficients in the polynomials means over
\\ GF(2)[x], ie. all coefficients are calculated mod 2.
\\
for(n=0,256, T(n) == hammingweight(Mod(1+x+x^2,2)^n)  || error());

\\ The whole row of bits can be formed, for successive rows, by suitable
\\ shifts and bitxor.  Such shifts and xor are the same as multiply by
\\ (1+x+x^2) in GF(2)[x].  This bitwise approach is good for iterating, but
\\ powering like the polynomial ought to be more efficient to find a
\\ particular individual row, for big n at least.

A038184(n) =
{
  my(b=1);
  for(i=1,n,
     \\ xors same as multiply 1+x+x^2 over GF(2)[x]
     b = bitxor(b,
                bitxor(b<<1,
                       b<<2)));
  b;
}
for(n=0,256, T(n) == hammingweight(A038184(n))  || error());

{
  print("Cellular Rule 150             num 1s = T(n)");
  for(n=0,8,
     my(b=A038184(n));
     printf("  n=%d%*d%*d\n",
            n,
            14+n, to_binary(b),
            15-n, hammingweight(b));
     hammingweight(b) == T(n) || error());
}

print();


\\-----------------------------------------------------------------------------
\\ a(n) Location in T
\\
\\ Since T contains all possible products of Chi(n) = J(n+2), it can be
\\ asked where the matchings count a(n) first appears in T.  The a(n)
\\ product of Js means the run lengths needed are a list
\\
\\     length     repeat
\\       n-1        1
\\       n-2        1
\\       n-3        2
\\       n-4        4
\\       ...       ...
\\        1       2^(n-2)
\\
\\ These lengths are for Chi, so for example the high term J(n+1) in a(n) is
\\ Chi(n-1) and so run length n-1.  J(0)=1 and J(1)=1 so multiplying those
\\ terms is no change and are omitted.

\\ return a repeating vector [value, value, ...] of length len
vector_repeats(len,value) = vector(len,i, value);

\\ return a vector [runlength, runlength, ...] needed to make a(n)
a_run_lengths(n) =
{
  my(ret=[]);
  for(i=1,n-2, ret=concat(ret, vector_repeats(2^(n-2-i), i)));
  if(n>=2, ret=concat(ret,[n-1]));
  ret;
}
{
  a_run_lengths(0) == []                        || error();
  a_run_lengths(1) == []                        || error();
  a_run_lengths(2) == [                 1]      || error();
  a_run_lengths(3) == [              1, 2]      || error();
  a_run_lengths(4) == [         1,1, 2, 3]      || error();
  a_run_lengths(5) == [1,1,1,1, 2,2, 3, 4]      || error();

  \\ products implied by the lengths
  for(n=0,12,
     my(v=a_run_lengths(n));
     prod(i=1,#v, Chi(v[i])) == a(n)  || error());
}

\\ The smallest t where T(t) = a(n) will be by taking a_run_lengths() with a
\\ single 0 bit separating each run, and taking the shortest runs in most
\\ significant position so the 0-bits are as high as possible.

\\ vec is a vector of run lengths
\\ return an integer t with 1-bit runs per vec, in the order of vec
run_lengths_to_index(vec) =
{
  for(i=1,#vec, vec[i] = concat(0, vector_repeats(vec[i],1)));
  if(#vec,vec=concat(vec));
  fromdigits(vec,2);
}
a_index_T(n) = run_lengths_to_index(a_run_lengths(n));

{
  a_index_T(2) == from_binary(1)                       || error();
  T(1) == a(2)                                         || error();

  a_index_T(3) == from_binary(1011)                    || error();
  a_index_T(4) == from_binary(1010110111)              || error();
  a_index_T(5) == from_binary(1010101011011011101111)  || error();
  \\                          1 1 1 1  2  2   3    4
}

{
  print1("a_index_T(n) = ");
  for(n=0,5,print1(a_index_T(n),", "));
  print();

  for(n=0,5,
     my(v=a_run_lengths(n),
        t=a_index_T(n));
     printf("  n=%d  %-25s  %d\n", n,v,to_binary(t)));
}
{
  for(n=0,12,
     my(t=a_index_T(n));
     T(t) == a(n)  || error());
}


\\-----------------------------------------------------------------------------
\\ a_index_T() Bit Length

\\ The bit length of a_index_T(n) is the total of its run lengths plus 1
\\ zero bit between each.
\\
\\     a_index_T_bitlength(n) = sum(a_run_lengths) + #a_run_lengths - 1
\\
\\       for n > 1 so that a_run_lengths not empty
\\
\\ The pattern of doublings except the two high terms in the Jacobsthals is
\\
\\     a_index_T_bitlength(n) = 2*a_index_T_bitlength(n-1) + 2   n>=3
\\                            = 3*2^(n-2) - 2      (OEIS A296953)
\\
\\       starting a_index_T_bitlength(2) = 1

a_index_T_bitlength(n) = if(n<2,0, 3*2^(n-2) - 2);
{
  print1("a_index_T_bitlength(n) = ");
  for(n=0,6,print1(a_index_T_bitlength(n),", "));
  print("...  (A296953)");
}
{
  a_index_T_bitlength(0) == 0   || error();
  a_index_T_bitlength(1) == 0   || error();
  a_index_T_bitlength(2) == 1   || error();
  a_run_lengths(2)      == [1]  || error();
  for(n=0,10,
     my(v=a_run_lengths(n));
     a_index_T_bitlength(n) == vecsum(v) + if(#v, #v-1)  || error());
  for(n=3,20,
     a_index_T_bitlength(n) == 2*a_index_T_bitlength(n-1) + 2  || error());
}

A296953(n) =
{
  n>=0 || error("A296953() is for n>=0");
  if(n==0,0, 3*2^(n-1) - 2);
}
{
  my(A296953_samples=[0,1,4,10,22,46,94,190,382,766,1534,3070,6142]);
  for(i=1,#A296953_samples,
     my(n=i-1); \\ A296953_samples OFFSET=0
     A296953(n) == A296953_samples[i]  || error());

  for(n=1,20,
     a_index_T_bitlength(n) == A296953(n-1)  || error());
}

print();


\\-----------------------------------------------------------------------------
\\ T Expected Growth

\\ T(n) values don't have an aymptotic limit as such, since T oscillates
\\ variously T(2^k) = 3 at every power-of-2 (single 1-bit Chi(1)=3), up to
\\ T(2^k-1) = Chi(k) which is a run of k 1-bits.
\\
\\ Some forms of "average" growth of T(n) are considered in
\\
\\     Steven Finch, Pascal Sebah and Zai-Qiao Bai, "Odd Entries in Pascal's
\\     Trinomial Triangle", February 2008
\\
\\ They take a uniformly distributed random N in the range 0 <= N < n and
\\ show the expected value of log(T(N)) is
\\
\\     expect log(T(N))  ~  log2(C) * log(n)
\\
\\ in the sense that the mean log(T(N)), divided by log(n), has a limit
\\
\\             sum(i=0,n-1, log(T(i)))     1                  log(C)
\\     limit   ----------------------- * ------   = log2(C) = ------
\\     n->inf             n              log(n)               log(2)
\\
\\ Roughly speaking, Chi(i) is about i bits long, and T(n) is the sum of
\\ those lengths, or thereabouts, so that this limit says something about
\\ runs of 1s on average comprising some proportion of the bits within N
\\ values < n.


\\ expected value of log(T(N)) for 0 <= N < n
ElogT(n) = sum(i=0,n-1, log(T(i))) / n;

{
  my(v=[5, 50, 200, 500, 1000, 5000]);
  for(i=1,#v,
     my(n=v[i]);
     printf("ElogT/log  n=%4d   %.6f...\n", n, ElogT(n)/log(n)));
  printf("    limit -> log2(C) = approx %.6f...\n", log2(C));
}

\\ Kropf and Wagner (reference below) consider the variance (sigma^2) of
\\ such log(T(N)).

print();


\\-----------------------------------------------------------------------------
\\ T Cumulative
\\
\\ The sum of the first n terms of T() is the total number of ON cells in
\\ the first n rows of the rule 150 automaton.
\\
\\                      n-1
\\     Tcumulative(n) = sum  T(i)           (OEIS A134659)
\\                      i=0
\\
Tcumulative_by_sum(n) = sum(i=0,n-1, T(i));
{
  print1("Tcumulative(n) = ");
  for(n=0,8,print1(Tcumulative_by_sum(n),", "));
  print("...  (A134659)");
}

\\ Tcumulative is OEIS sequence A134659, except the sequence sums 0 to n
\\ inclusive, which is n+1 terms, instead of 0 to n-1 which is n terms here.
\\ Taking n terms is more convenient for the blocks forms below.
{
  my(A134659_samples=[1,4,7,12,15,24,29,40,43,52,61,76,81,96,107,128,131]);
  vector(#A134659_samples,n,n--; Tcumulative_by_sum(n+1)) == A134659_samples
    || error();    \\ A134659_samples OFFSET=0
}

\\ Tcumulative of the first 2^k many terms is
\\
\\     Tblock2k(k) = Tcumulative(2^k)
\\                 = 2^k*Fibonacci(k+2)            (OEIS A087206)
\\
\\ as per
\\
\\     Jens Christian Claussen, "Time-Evolution of the Rule 150 Cellular
\\     Automaton Activity From a Fibonacci Iteration", October 2004,
\\     section 5.  http://arxiv.org/abs/math/0410429
\\
\\ who makes a geometric construction similar to Wolfram's figure 6 (page
\\ 607 or authors' page 13) for the following recurrence and consequent
\\ Fibonacci form.
\\
\\     Tblock2k(k) = 2*Tblock2k(k-1) + 4*Tblock2k(k-2)    starting 1,4
\\
\\ Wolfram (page 616 or author's page 26) has this sort of recurrence
\\ expressed for triangle density and from which the fractal dimension of
\\ the limit set of points in rows 0 to 2^k-1 as k -> infinity.

Tblock2k(k) = 2^k * fibonacci(k+2);  \\ per Claussen

{
  print1("Tblock2k(n) = ");
  for(n=0,7,print1(Tblock2k(n),", "));
  print("...  (A087206)");
}
{
  for(k=0,8, Tblock2k(k) == Tcumulative_by_sum(2^k)             || error());
  for(k=2,10, Tblock2k(k) == 2*Tblock2k(k-1) + 4*Tblock2k(k-2)  || error());

  my(A087206_samples=[1,4,12,40,128,416,1344,4352,14080,45568,147456,477184]);
  vector(#A087206_samples,k,k--; Tblock2k(k)) == A087206_samples
    || error();    \\ A087206_samples OFFSET=0
}

\\ Tblock2k allows the total Tcumulative to be formed by blocks.
\\ At each 1-bit in n, take T(i) for those i of the form
\\
\\              1 bit in n ---v  |-- k bits --|
\\         +----------------+---+--------------+
\\     i = | high bits of n | 0 |     ...      |
\\         +----------------+---+--------------+
\\            T(high bits)    *    Tblock2k(k)
\\
\\ These i values are all < n because at bit position k (counting the least
\\ significant bit as position 0) the 1-bit in n is a 0-bit instead.
\\ The low bits shown "..." are arbitrary.  Total Tcumulative() of those low
\\ values is Tblock2k(k), and then the runs in the "high bits of n" are a
\\ run length transform product.

Tcumulative_by_blocks(n) =
{
  sum(k=0,bitlength(n)-1,
      if(bittest(n,k), T(n>>(k+1)) * Tblock2k(k)));
}
{
  for(n=0,128,
     Tcumulative_by_blocks(n) == Tcumulative_by_sum(n)  || error());
}

\\ This is a fairly usual sort of block breakdown for a common high part and
\\ a range of low.  A general treatment can be found in
\\
\\     Sara Kropf, Stephan Wagner, "q-Quasiadditive Functions", Proceedings
\\     of the 27th International Conference on Probabilistic, Combinatorial
\\     and Asymptotic Methods for the Analysis of Algorithms, Krakow, Poland,
\\     4-8 July 2016.  http://arxiv.org/abs/1605.03654
\\
\\ Their example 8 at the end of the paper is t_n = T(n).  They note T(n) is
\\ 2-quasimultiplicative in their sense with their "r=1",
\\
\\     T(2^(k+1)*x + y) = T(x)*T(y)         for 0 <= y < 2^k
\\
\\ which here is due to the 0-bit at position k separating the runs in the
\\ high part x and the low part y.
{
  for(k=0,6,
     for(x=1,12,
        for(y=0,2^k-1,
           T(2^(k+1)*x + y) == T(x)*T(y)  || error())));
}
\\
\\ Kropf and Wagner consider uniform random log(T(n)) in a similar way to
\\ Finch et al and reach the same expected value.  At theorem 14 they write
\\ constant "mu" in mean mu*k, with k = log2(n) for the case q=2.  This
\\ means log2(n) instead of log(n) in the expectations set out above, and so
\\ the constant mu = log(C) instead of log2(C).


\\ In Tcumulative_by_blocks(), the T(high bits) factors can be formed by
\\ accumulating the products for run lengths as the bits of n are considered
\\ successively, high to low.  In the following code, Thigh = T(high bits)
\\ for completed high run 1s.  len!=0 is the length of an in-progress run of
\\ 1s.
\\
\\ In the bittest==0 case, have Chi(0) = 1 so that Thigh *= Chi(len) could
\\ be unconditional, but doing it when len!=0 illustrates the end-of-run
\\ accumulation of a completed run into Thigh.
\\
Tcumulative_by_progressive(n) =
{
  my(len=0, ret=0, Thigh=1);
  forstep(k=bitlength(n)-1,0,-1,
     if(bittest(n,k),
        ret += Thigh * Chi(len) * Tblock2k(k);
        len++;
        ,
        if(len, Thigh *= Chi(len);
        len=0)));
  ret;
}
{
  for(n=0,256,
     Tcumulative_by_progressive(n) == Tcumulative_by_blocks(n) || error());
  Chi(0) == 1  || error();
}

\\ Tblock2k() calls fibonacci(k+2), which is various squarings and
\\ additions.  If n has many 1-bits then that will be many fibonacci()
\\ calls, often to nearby k.  An optimization can be made by calculating all
\\ needed Tblock2k together using polynomial evaluation at an x = t_POLMOD.
\\ The polynomial terms summed into ret are like
\\
\\     Tcumulative(n) = v[1]*x^k + ... + v[#v]*x^0
\\
\\ where the v[] coefficients are the Thigh*Chi(len) values (integers), and
\\ x is a polmod which, when subst()ed into this polynomial, and thereby
\\ powered up, gives x^k = Tblock2k(k).
\\
\\ Circa Pari version 2.9.x, GP level subst(), which is library poleval(),
\\ is only a Horner type polynomial evaluation, but it recognises sparse
\\ coefficients and skips runs of zeros by a power x^(how many 0 coeffs).
\\ So a sparse (few 1-bits) n should still be roughly equal in speed to
\\ Tcumulative_by_progressive().

Fib2k_m = Mod('x, 'x^2 - 2*'x - 4);

Tcumulative(n) =
{
  my(v=binary(n), len=0, high=1);
  for(i=1,#v,
     if(v[i], v[i] = high*Chi(len); len++,
              if(len, high *= Chi(len)); len=0));
  v = lift(subst(Pol(v),'x, Fib2k_m));
  polcoeff(v,0) + 4*polcoeff(v,1);
}
{
  for(n=0,512,
     Tcumulative(n) == Tcumulative_by_blocks(n) || error());
}

print();

{
  \\ Here's a sample measure to illustrate speed of polmod vs separate
  \\ Tblock2k.  It also illustrates greater stack memory use by polmod.
  \\
  \\ Don't want to run a speed test like this every time in an example
  \\ program, and anyway the length or repetitions needed to show a
  \\ difference will depend on how fast or slow your computer or Pari build,
  \\ now or in the future.  Enable by "if(1," and adjust k up or down to
  \\ provoke.
  \\
  \\ At about k=16000, the polmod speedup is about 5*.  That's a 16000 bit
  \\ number n, which might seem an improbable size, but it's one of those
  \\ things where a good algorithm opens up the chance to go large should
  \\ that be desired.
  \\
  if(0,
     my(k = 16000,
        n=2^k + random(2^k));

     gettime(); Tcumulative_by_progressive(n);
     print("Tcumulative_by_progressive millseconds  ",gettime());

     gettime(); Tcumulative(n);
     print("Tcumulative                millseconds  ",gettime());
     );
}


\\-----------------------------------------------------------------------------
\\ Tcumulative Expectation
\\
\\ Finch, Sebah and Bai consider another expected growth of T.  They show
\\ for a random N in the range 0 <= N < n, the expectation (mean) is
\\
\\     expect T(N)  ~  n^log2(phi)
\\
\\     where golden ratio  phi = (1+sqrt(5))/2       (OEIS A001622)
\\     log2(phi) = 0.6942419...                      (OEIS A242208)
\\
\\ in the sense that
\\
\\         log (expect T(N))          log( Tcumulative(n) / n )
\\   lim   -----------------  = lim   -------------------------  = log2(phi)
\\   n->inf     log(n)          n->inf        log(n)

ET(n) = Tcumulative(n) / n;    \\ expected T(N) for 0 <= N < n
phi = (1+sqrt(5))/2;
{
  my(v=[10,100,1000,10000]);
  for(i=1,#v,
     my(n=v[i]);
     printf("ET/log  n=%5d   %.6f...\n", n, log(ET(n))/log(n)));
  printf("    cf log2(phi) = approx %.7f...     (A242208)\n",log2(phi));
}

\\ The block strategy for Tcumulative allows some big sample values to be
\\ tried.  The most obvious example for the limit is n = 2^k which is
\\ Tblock2k
\\
\\     ET(2^k) = 2^k * Fibonacci(k+2) / 2^k
\\             = Fibonacci(k+2)
\\
\\ and with Binet's formula for Fibonacci in terms of phi,
\\
\\     log(ET(2^k)) / log(2^k) ~ log(phi^(k+2)/sqrt(5)) / log(2^k)
\\                             = (k*log(phi) + log(phi^2/sqrt(5))) / (k*log(2))
\\                             = log2(phi)  + log2(phi^2/sqrt(5))/k
\\
\\ Term log2(phi^2/sqrt(5)) / k decreaes only as 1/k so convergence won't be
\\ particularly fast.
\\
\\ Believe Wolfram's density limit (equation 3.6 page 616 and author's copy
\\ page 26) effectively has this.  Or does it go further there than only
\\ n = 2^k ?

{
  for(k=200,510,
     my(n=2^k);
     ET(n) == fibonacci(k+2)                                || error();
     nearly_equal(fibonacci(k+2) / (phi^(k+2)/sqrt(5)),  1) || error();
     nearly_equal(log(ET(n))/log(n),
                  (k*log(phi) + log(phi^2/sqrt(5))) / (k*log(2))  ) || error();
     nearly_equal(log(ET(n))/log(n),
                  log2(phi)  + log2(phi^2/sqrt(5)) / k )        || error();
  );
}


\\-----------------------------------------------------------------------------
\\ T() Bits
\\
\\ For further interest, some of the low bits of T(n) values can be
\\ characterized.
\\
\\ T(n) is odd since it is the product of Chi terms which are all odd,
{
  vector(100,n, bittest(Chi(n),0)) == vector(100,n, 1)  || error();
  vector(100,n, bittest(T(n),  0)) == vector(100,n, 1)  || error();
}

\\ Bit position 1 (second lowest bit) of Chi is 0 or 1 according as n even
\\ or odd, for n>=1.  Together with Chi bit position 0 always 1-bit means
\\
\\     Chi(n) = 1 or 3 mod 4   according as n even or odd
\\
\\ Products of Chi for T(n) are then T(n) = 1 or 3 mod 4 according as the
\\ sum of the 1-bit run lengths in n is even or odd.  That sum is the total
\\ number of 1 bits in n and its parity is the Thue-Morse sequence.

ThueMorse(n) = hammingweight(n)%2;   \\ OEIS A010060
{
  vector(100,n, Chi(n)%4)        == vector(100,n, if(n%2==0,1,3))  || error();
  vector(100,n, bittest(T(n),1)) == vector(100,n, ThueMorse(n))    || error();
  vector(100,n, T(n)%4)          == vector(100,n, if(ThueMorse(n)==0, 1,3))
                                                                   || error();
}
\\ Thue-Morse as values 1,3, which is T(n) mod 4, might be OEIS A268676.
\\ Antti Karttunen notices this in that sequence.  Very likely, but the
\\ sequence definition is multiple Gray code and UnGray code steps which
\\ would need working through.

\\ T(n) bit position 2 (third lowest) goes as the Golay-Rudin-Shapiro
\\ sequence, as per Ralf Stephan's formula in A071053.
\\
\\ Chi bit position 2 is 1 or 0 according as n even or odd, for n>=1.
\\ This and the other low bits of Chi mean
\\ 
\\      Chi(n) = 5 or 3 mod 8  according as n even or odd
\\
\\ The multiplication table for odd integers mod 8 is
\\
\\         1  3   5  7
\\       +-------------
\\     1 | 1  3   5  7
\\     3 | 3  1   7  5
\\       |
\\     5 | 5  7   1  3
\\     7 | 7  5   3  1
\\
\\ Notice that multiplying 5 or 7 toggles between 1or3 and 5or7 whereas
\\ multiplying 1 or 3 remains the same 1or3 or 5or7.  So bit 2 in T(n)
\\ toggles when an even-length run of 1-bits in n, so parity of the number
\\ of even-length runs.  Such parity is the Golay-Rudin-Shapiro sequence.
\\ GRS is usually reckoned with values +1,-1 as (-1)^parity.
\\
\\ Function GRS01 below is the GRS expressed in the form of 0 or 1 according
\\ as parity of the number of 11 bit pairs in n, with overlapping pairs
\\ allowed.  bitand(n,n>>1) keeps 1-bits with another 1 immediately above.
\\
\\ Counting 11 bit pairs for GRS is shown by Brillhart and Carlitz and is
\\ often used as a definition for GRS in fact.  The equivalence to parity of
\\ even length runs is because an odd length run of 1s is an even number of
\\ 11 pairs, and an even length run of 1s is an odd number of 11 pairs.

GRS01(n) = hammingweight(bitand(n,n>>1)) % 2;
{
  vector(100,n, Chi(n)%8)        == vector(100,n, if(n%2==0,5,3))  || error();

  \\ per Ralf Stephan
  vector(100,n, bittest(T(n),2)) == vector(100,n, GRS01(n))        || error();

  \\ product -1 for each even length run of 1-bits in n
  vector(100,n, (-1)^GRS01(n)) ==
  vector(100,n, run_length_transform_of_func(n,
                                    len->(-1)^(len-1)))            || error();
}


\\-----------------------------------------------------------------------------
print("end");
