\\ Copyright 2018, 2019, 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0);
default(strictargs,1)
read("../vpar.gp");


\\ Usage: gp <integral-tree-diameter.gp
\\
\\ An integral tree (or graph) has integers for all eigenvalues of its
\\ adjacency matrix (vpar_adjacency_matrix()).  Or equivalently, all
\\ integers for roots of the characteristic polynomial of that matrix.
\\ In vpar.gp that polynomial is vpar_adjacency_poly(vpar).
\\
\\ Csikvari constructs integral trees of specified diameter.
\\
\\     Peter Csikvari, "Integral Trees of Arbitrarily Large Diameters",
\\     J. Algebr. Combin., volume 32, 2010, pages 371-377.
\\
\\ make_T() below takes r1..rk, and make_S() takes desired roots S =
\\ [n1,...,nk].
\\
\\ The sample run is Csikvari's example spectrum S=1,2,4,5 from a tree of
\\ 781 vertices.
\\
\\ Csikvari's further example spectrum S=1,2,3,4,5,6 of 27007 vertices can
\\ be constructed quickly.  vpar_adjacency_poly() on it is just within reach
\\ of the code but it takes some time due to intermediate polys of growing
\\ size and coefficients which soon become bignums.  Expect peak about 350Mb
\\ stack in 32-bits (though final result only about 20Mb), in about 10
\\ minutes of 6000 bogomips.  Checking it with factor() or even making and
\\ comparing factorback() uses even more stack.  Checking and dividing each
\\ expected factor keeps stack down but is slow (over an hour).
\\


\\-----------------------------------------------------------------------------

\\ Return the vpar tree of vec = [r1, r2, ..., rk].
\\
\\ "this" is vertex numbers of A_i and "prev" is vertex numbers of A_{i-1}.
\\ r_i many new vertices are attached onto each of A_i.  They and A_{i-1}
\\ become the new A_{i+1} vertices.
\\
\\ The code here keeps these vertices in lists.  They are contiguous ranges
\\ so just the starting number of each block would be more efficient.
\\
make_T(vec) =
{
  my(vpar=[0],
     this=List([1]),
     prev=List([]));
  for(i=1,#vec,
     my(r=vec[i],
        upto=#vpar);
     vpar=Vec(vpar, #vpar + r*#this);   \\ extend
     for(t=1,#this,
        for(j=1,r,
           vpar[upto++] = this[t];
           listput(prev,upto)));
     my(temp=this); this=prev; prev=temp);
  vpar;
}

\\ r_{k-i} = n_{i+1}^2 - n_i^2
\\
S_to_r(vec) = \
  vector(#vec,i, vec[#vec-i+1]^2 - if(i<#vec,vec[#vec-i]^2));

\\ Return the vpar tree of vec = [n1, n2, ..., nk].
\\ Must have ascending n1 < n2 < ... < nk.
\\ It has diameter 2*k and the roots of its characteristic polynomial are
\\ all and only n1, ... nk, and 0.
\\
make_S(vec) = make_T(S_to_r(vec));


\\-----------------------------------------------------------------------------
\\ Csikvari's example of 781 vertices

S=[1,2,4,5];
print("example S = "S);
print("r of S = "S_to_r(S));

vpar=make_S(S);
print("num vertices "#vpar);
print("factorize poly "factor(vpar_adjacency_poly(vpar)));


\\-----------------------------------------------------------------------------
\\ number of vertices in the tree of spectrum 1,...,n

print();
vec=vector(6,n,n--; #make_S(vector(n,i,i)));
print("spectrum 1..n number of vertices = "vec);
print("which is A001515 = Bessel polynomial y_n(x) at x=1");

\\-----------------------------------------------------------------------------
