\\ Copyright 2018, 2019, 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0);
default(strictargs,1)
read("../vpar.gp");


\\ Usage: gp <integral-tree-search.gp
\\
\\ This is a brute-force search for integral trees.
\\
\\ An integral tree (or graph) has an all integer spectrum, where spectrum
\\ means the eigenvalues of its adjacency matrix.  Or equivalently, all
\\ integer roots of its characteristic polynomial, that being characteristic
\\ polynomial of its adjacency matrix.
\\
\\ In vpar.gp, vpar_adjacency_matrix() is the matrix and
\\ vpar_adjacency_poly() is the characteristic polynomial.
\\
\\ ----------
\\ This type of search can be done much better.  For example
\\
\\     A. E. Brouwer, "Small Integral Trees", Electr. J. Combin., volume 15,
\\     2008
\\     https://www.win.tue.nl/~aeb/graphs/integral_trees.html 
\\     http://www.win.tue.nl/~aeb/preprints/small_itrees.pdf
\\
\\ Brouwer notes that Cauchy interlacing of matrix eigenvalues means that
\\ the eigenvalues of a graph G and of its delete one vertex G\x interlace.
\\ The adjacency matrices of any graph is real and symmetric so all roots
\\ are real.
\\
\\     G roots    g[1] <= g[2] <= ... <= g[n]      n vertices, n roots
\\     G\x roots  s[1] <= s[2] <= ... <= s[n-1]    n-1 vertices, n-1 roots
\\
\\     interlacing:
\\     
\\     g[1] <= s[1] <= g[2] <= s[2] <= ... <= s[n-1] <= g[n]
\\
\\ If all of g are to be integers then range s[1] to s[2] must contain an
\\ integer for g[2], and s[2] to s[3] for g[3], etc.
\\
\\ These s and g are inclusive of any duplication of roots their polynomials
\\ might have.  A duplicate non-integer root in G\x means G cannot be
\\ integral.
\\
\\ For a free tree G, deleting to G\x is a free forest.  If one of those
\\ component trees has a pair of eigenvalues without an integer between then
\\ G is not integral.  This means such a tree is never a branch of an
\\ integral tree.
\\
\\ The poly roots of a forest are the union of its component trees.
\\ Consecutive roots of G\x might come one from one component and the next
\\ from another.  If they have no integer between then such a pair of trees
\\ cannot be branches attached at the same place in an integral tree.
\\
\\ Brouwer made a search up to N=50 by truncating the search space with such
\\ considerations.  It can be noted this interlacing is only a necessary
\\ condition, not sufficient.  Just because G\x has room for integers
\\ between doesn't mean G will in fact be such integers.
\\
\\ ----------
\\
\\ is_integral_poly() below relies on factorizing.  This is fine for small
\\ to medium polynomials.  For big might prefer some polsturm() bisecting a
\\ range of possible roots and stopping if a non-integer found.  The lowest
\\ non-zero term of the poly is product its non-zero roots so that is a
\\ bound on the roots range.
\\
\\ The addhelp string of vpar_adjacency_poly() describes how it derives from
\\ the match polynomial.  Integer roots of vpar_adjacency_poly() are pairs
\\ +/-r which come from terms (r^2*x + 1) in vpar_matchpoly().  So an
\\ integral tree has matchpoly roots all 1/r^2.  Applying a reciprocal and
\\ negate subst(polrecip(vpar_matchpoly(vpar)),'x,-'x) becomes (x-r^2) so
\\ roots of that all integer squares.
\\


\\-----------------------------------------------------------------------------

\\ Return 1 if poly has integers for all its roots.
is_integral_poly(poly) =
{
  my(m=factor(poly));
  for(i=1,matsize(m)[1],
     if(poldegree(m[i,1])>1, return(0)));
  1;
}

limit = 14;
print("integral trees n=0 to "limit"    (OEIS A331414)");
{
  for(n=0,limit,
     my(vpar=vpar_free_first(n));
     until((vpar=vpar_free_next(vpar))=='none,
           my(poly=vpar_adjacency_poly(vpar));
           if(is_integral_poly(poly),
              print("n="n);
              print("  vpar="vpar);
              print("  factors "factor(poly));
             )));
}

\\-----------------------------------------------------------------------------
