\\ Copyright 2018, 2019, 2020, 2021 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(parisizemax, 100*10^6);  \\ 100 mbytes
default(recover,0);
default(strictargs,1)
read("../vpar.gp");


\\ Usage: gp <isomorphic-halves.gp
\\
\\ Requires GP 2.9 up for Map().
\\
\\ This is an experiment finding trees which are bicentral and have their
\\ two halves isomorphic as free trees, and those halves in turn are
\\ bicentral and free isomorphic halves, and so on down to each half a
\\ single vertex (which is path-2).
\\
\\ Such trees have 2^k many vertices, by their construction.  They can be
\\ made by connecting two copies of a 2^(k-1) tree at the same vertex in
\\ each, or at vertices of equal eccentricity (so as to maintain bicentral).
\\
\\ Examples of these trees include, all being "connect at same vertex" type,
\\
\\     * binomial tree (vpar_make_binomial_tree())
\\     * twindragon area tree
\\     * twin alternate area tree
\\
\\ Binomial tree connects two 2^(k-1) trees at their centres (either of the
\\ two bicentral vertices).  Twindragon and twin alternate connect at
\\ certain leaf vertices (see the author's Graph::Maker Perl code and docs,
\\ or long mathematical writeups for details).
\\
\\ The code here counts how many of this type of tree, up to free
\\ isomorphism.
\\
\\ The search is a moderately naive construction of all possible joinings,
\\ with de-duplication by going canonical vpar_to_free().  The vertices to
\\ attempt joining are reduced by considering only 1 from each free
\\ automorphism orbit.  If two vertices are in the same orbit then the rest
\\ of the graph looks the same at each (cf notes under "Orbits" in vpar.gp
\\ initial comments) and so give the same joined result.
\\
\\ A good amount of memory and a bit of time is saved by using Vecsmall for
\\ the vpar keys in "seen".  This is like described under "Vector Types" in
\\ the vpar.gp initial comments.  vpar_join() etc which are used currently
\\ accept Vecsmall for their operations, though that is not a guaranteed
\\ feature.
\\
\\ Results
\\ -------
\\
\\ Number of trees:
\\ Size 2^k vertices          k = 0  1  2  3  4    5     6      7
\\ Connect at same vertex:        1, 1, 1, 2, 8,  64, 1024, 32768  (A006125)
\\ Connect at equal eccentricity: 1, 1, 1, 2, 9, 122, 5894
\\
\\ The number of trees connect at same vertex each half is
\\
\\      S(k) = S(k-1) * 2^(k-1)/2     starting S(0)=S(1)=1
\\           = if(k==0,1, 2^((k-1)*(k-2)/2) )          A006125
\\
\\ This is any of the k-1 trees copied and joined at any of its 2^(k-2)
\\ vertices in one half of the k-1.
\\
\\ To see that all k made this way are different, firstly the only
\\ automorphism in these trees is a swap of halves.  This is so in k=1
\\ path-2 swapping its 2 vertices.  Then supposing it true in k-1, joining
\\ to form k causes the joined half of k-1 to be different from its other
\\ half, destroying its swap automorphism.  Consequently, all automorphism
\\ orbits are pairs of vertices, one vertex from one half and one vertex
\\ from the other.
\\
\\ Two k trees made from different k-1 halves are different, since an
\\ isomorphism between the two k would be an isomorphism between their halves.
\\ Two k trees made from the same k-1 but joined at vertices from different
\\ orbits are different, since any isomorphism between the k would be an
\\ automorphism of k-1 which mapped different orbit vertices to each other.
\\
\\
\\ GP-DEFINE  S(k) = if(k==0,1, 2^((k-1)*(k-2)/2) )
\\ GP-Test  vector(6,k,k++; S(k)) == vector(6,k,k++; S(k-1) * 2^(k-1)/2)
\\ vector(6,k,k--; S(k))
\\ not in OEIS: 1, 2, 9, 122, 5894     \\ at equal eccentricity


print("Connected at same vertex:");
pending = [ [0] ];  \\ singleton
print("k=0 size "#pending[1]" count "#pending);
{
  for(k=1,6,
     my(seen=Map());
     for(i=1,#pending,
        my(vpar=(pending[i]),
           orbits=vpar_free_orbit_sets(vpar));
        #orbits == ceil(#vpar/2) || error();
        for(j=1,#orbits,
           my(v=orbits[j][1],                 \\ first of each orbit
              new_vpar=vpar_join(vpar,v, vpar_reroot(vpar,v)));
           new_vpar=vpar_to_free(new_vpar);   \\ canonical
           new_vpar=Vecsmall(new_vpar);       \\ save memory
           mapput(seen,new_vpar,0)));
     pending=Mat(seen)[,1];
     print("k="k" size "#pending[1]" count "#pending));
}

print();

print("Connected at equal eccentricity:");
pending = [ [0] ];  \\ singleton
print("k=0 size "#pending[1]" count "#pending);
{
  for(k=1,6,
     my(seen=Map());
     for(i=1,#pending,
        my(vpar=Vec(pending[i]),
           orbits=vpar_free_orbit_sets(vpar),
           eccentricities=vpar_eccentricities_vector(vpar));
        for(j=1,#orbits,
           my(v=orbits[j][1]);                \\ first of each orbit
           for(k=1,j,
              my(u=orbits[k][1]);             \\ first of each orbit
              if(eccentricities[u]==eccentricities[v],
                 my(new_vpar=vpar_join(vpar,u, vpar_reroot(vpar,v)));
                 new_vpar=vpar_to_free(new_vpar);   \\ canonical
                 new_vpar=Vecsmall(new_vpar);       \\ save memory
                 mapput(seen,new_vpar,0)))));
     pending=Mat(seen)[,1];
     print("k="k" size "#pending[1]" count "#pending));
}
