\\ Copyright 2019, 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0)
default(strictargs,1);
read("../vpar.gp");


\\ Usage: gp <random-decorated.gp
\\
\\ This is some code generating a random decorated binary tree, and
\\ corresponding ordered forest, by the method of
\\
\\     Jean-Luc Remy, "Un Procede Iteratif de Denombrement d'Arbres Binaires
\\     et Son Application a Leur Generation Aleatoire", RAIRO Informatique
\\     Theorique, volume 19, 1985, pages 179-195.
\\     http://www.numdam.org/item/ITA_1985__19_2_179_0/
\\     http://www.numdam.org/article/ITA_1985__19_2_179_0.pdf
\\
\\ Knuth gives this in volume 4A section 7.2.1.6 algorithm R, page 16 of
\\ draft http://www-cs-faculty.stanford.edu/~knuth/fasc4a.ps.gz
\\
\\ A binary tree, in the usual way, has "internal" vertices with left and
\\ right child at each.  Empty child positions are reckoned as "external".
\\ For n internal vertices, there are n+1 externals.  The "decoration" of
\\ the tree is n+1 labels at the external positions.  For example,
\\
\\              *
\\             / \        * = internal vertex (no label)
\\            *   C       A,B,C = external vertex positions, labelled
\\           / \
\\          B   A
\\
\\ The program code below is Knuth's algorithm R making a linked binary tree
\\ in a single vector.  But here vertex numbers are adapted to 1-based
\\ indexing used by GP vectors.
\\
\\ A tree of n internal vertices has 2n+1 total internal+external.  They are
\\ distinguished by odd,even so
\\
\\     odd v  = external  1,3,...,2n+1   no children
\\     even v = internal  2,4,...,2n     left and right child positions
\\
\\ For v even, its children are recorded in vector L at
\\
\\     L[v]   = left child
\\     L[v+1] = right child
\\
\\ Pairs start L[2],L[3] are the children of v=2, then L[4],L[5] are the
\\ children of v=4, etc.  Such a pair is the usual binary tree "cell" (like
\\ a Lisp cons cell), here in consecutive vector entries.  L[1] is the root
\\ vertex number.  For n=0 empty tree, L[1]=1 for an external vertex
\\ "underneath" an empty tree.  Otherwise L[1] is even for an internal vertex.
\\
\\ For example, n=2 internal vertices,
\\
\\          4              v---root vertex (an internal)
\\         / \        L = [4, 3,1, 2,5]
\\        2   5               ^    ^---- left,right children of 4
\\       / \                  +--- left,right children of 2
\\      3   1
\\
\\    2,4 internal
\\    1,3,5 external
\\
\\ For each tree structure, there are n+1 external positions and the random
\\ choice gives external labels 1,3,5,... in all (n+1)! arrangements at
\\ those positions.
\\
\\ The algorithm is to choose a random vertex k among the 2n+1 internals +
\\ externals, and insert there a new internal i=even so
\\
\\     k   =>      i              i
\\                / \      or    / \
\\              i+1  L[k]     L[k]  i+1
\\
\\ In terms of L vector, k is one of the positions, so one of the L[1] root
\\ or L[2,3] etc cell.  Changing L[k]=i sets i as the child in that place,
\\ or sets i as the root.  This new i is even and its children L[i],L[i+1]
\\ are set to a new i+1 external (odd), and to L[k] which was the previous
\\ contents of that link pushed downwards.  These i+1,L[k] are either way
\\ left,right or right,left by random choice.
\\
\\ In the output, the internal vertex labels are not permuted or otherwise
\\ have any choice, rather they are determined by the external labels.
\\ When i is inserted, it gets odd o=i+1 immediately below it.  An internal
\\ with 2 external children occurs only by an insert applied to a smaller
\\ external.  So internal has 2 external o1 and o2 then is i = max(o1,o2)-1.
\\ The smaller min(o1,o2) can replace i and repeat until all internals are
\\ labelled.
\\
\\ Per Remy, and as set out in Knuth, the random choice is 2*(2n+1) on
\\ extending decorated n to decorated n+1.  This is directly in the count of
\\ decorated trees.  There are C(n) binary trees
\\
\\     C(n) = binomial(2n,n) / (n+1)  \\ Catalan, num binary trees n internal
\\
\\ and decoration (n+1)! so
\\
\\     D(n) = (n+1)! * C(n)    \\ num decorated binary trees n internal
\\          = (2n)! / n!
\\          = 1, 2, 12, 120, 1680, 30240, 665280, ...   (A001813)
\\
\\ In the usual way Catalan C(n+1) = 2*(2*n+1)/(n+2) * C(n) and consequently
\\ factor 2*(2n+1) on D(n)
\\
\\     D(n+1) = 2*(2n+1) * D(n)

C(n) = vpar_count_ordered(n,1);
vector(10,n,n--; C(n+1)) ==  /* stepping C(n) */   \
vector(10,n,n--; 2*(2*n+1)/(n+2) * C(n))  || error();

D(n) = (n+1)! * C(n);
vector(10,n,n--; D(n+1)) ==  /* stepping D(n) */   \
vector(10,n,n--; 2*(2*n+1) * D(n))  || error();


\\ Return an L vector which is a uniform random choice of decorated binary
\\ tree.
random_decorated_L(n) =
{
  my(L=vector(2*n+1));
  L[1]=1;
  forstep(i=2,2*n,2,
     my(k=random(i-1)+1,   \\ preceding internal or external
        b=random(2));      \\ bit which way around children of i
     L[i+b]  = i+1;        \\ child new external
     L[i+!b] = L[k];       \\ child previous at k
     L[k] = i);            \\ at k is now internal i
  L;
}

\\ L_to_vpar() takes an L vector and returns the preorder vpar forest
\\ resulting from the natural correspondence of binary tree to ordered
\\ forest.  The decorations of external vertices in L are ignored.
\\
\\ This code is a simple but inefficient recursion.  If v is an external
\\ vertex (odd) then return empty.  If v is an internal vertex (even) then
\\ make a root [0] and put left subtree below and right subtree concat as
\\ next.  The default v=L[1] begins at the binary root.
\\
\\          root ... right-subtree
\\           |
\\          left-subtree
\\
L_to_vpar(L, v=L[1]) =
{
  if(v%2, [],
     vpar_concat_forest(vpar_join([0],1, L_to_vpar(L,L[v])),
                                         L_to_vpar(L,L[v+1])));
}

\\ L_to_vpar_by_traverse() is the same as L_to_vpar() but implemented by
\\ traversing the binary tree.  This is more efficient, but not so clear.
\\
\\ At an internal vertex, descend to the left child, and push the vertex
\\ onto a linked list "chain" ready to descend into the right child later.
\\ chain = an i (even) internal.  L[i] has been changed to be the next
\\ preceding entry of the chain, and so on until 0 end of chain.
\\ 
\\ An internal is a new vertex v for vpar.  p is its parent.
\\ On descending to left child, p is the new vertex v.
\\ On backtracking up the chain, p goes to its parent vpar[p].
\\ This maintaining of p is similar to what vpar_make_preorder_depths()
\\ does, but there "backtracking" goes by next depths[v] <= previous.
\\
L_to_vpar_by_traverse(L) =
{
  my(i=L[1],
     vpar=vector(#L>>1),
     chain=0,
     v=1,
     p=0);
  while(1,
        \\ print("at i="i" v="v" chain="chain"  "L"  "vpar);
        if(bitand(i,1),
           \\ print(" external, backtrack");
           if(!chain, break);     \\ at top again
           i=L[chain+1];          \\ right leg
           chain=L[chain];        \\ pop
           p=vpar[p];
           ,
           \\ print(" internal, descend left leg");
           my(left=L[i]);
           L[i]=chain; chain=i;   \\ push
           i=left;
           vpar[v]=p;
           p=v;
           v++);
       );
  vpar;
}


\\ print L with pairs left,right grouped together
print_L(L) =
{
  print1("["L[1]);
  forstep(i=2,#L-1,2,
     print1(", "L[i]","L[i+1]));
  print("]");
}

{
  my(L=random_decorated_L(3),
     vpar=L_to_vpar(L));
  print1("random L  "); print_L(L);
  print("  vpar    "vpar);
  L_to_vpar_by_traverse(L) == vpar   || error();
}


\\-----------------------------------------------------------------------------
\\ my_random(limit) is like random(limit), but takes its values from
\\ my_random_list.  It returns my_random_list[1] and shifts that off so next
\\ time will be what was my_random_list[2], etc.
\\ my_random_limit_list is a similar list, but of the "limit" parameter to
\\ expect on each call, used for a consistency check that the my_random()
\\ call matches the intended value about to be supplied.
\\
my_random_list = List([]);
my_random_limit_list=List([]);
my_random(limit) =
{
  my(r=my_random_list[1]);                 listpop(my_random_list,1);
  my(want_limit=my_random_limit_list[1]);  listpop(my_random_limit_list,1);
  limit==want_limit    || error("oops");
  (0<=r && r<limit)    || error("oops");
  r;
}

\\ Run the random_decorated_L() code with the random() there replaced by a
\\ loop across all possible values at each random call, thereby producing 1
\\ each of all the L vectors random_decorated_L() would produce.
\\ The return is a vector of all other L vectors.
\\
random_decorated_all(n) =
{
  my(limits=vector(2*n,i, if(i%2,i,2)),
     ret=List([]));
  forvec(values=apply(limit->[0,limit-1], limits),
        my_random_list=List(values);
        my_random_limit_list=List(limits);

        my(L=vector(2*n+1));
        L[1]=1;
        forstep(i=2,2*n,2,
           my(k=my_random(i-1)+1,
              b=my_random(2));
           L[i+b]  = i+1;
           L[i+!b] = L[k];
           L[k] = i);

        listput(ret,L));
  Vec(ret);
}

\\ vec is a vector.  Return the number of occurrences of elem in vec.
vector_count_occurrences(vec,elem) = sum(i=1,#vec, vec[i]==elem);

{
  \\ Check that random_decorated_L() and L_to_vpar() gives a uniformly
  \\ distributed pre-order vpar, by checking how many random_decorated_all()
  \\ total, and how many of each distinct vpar.
  \\
  \\ Each different preorder vpar forest should occur.  There are
  \\ vpar_count_ordered(n,1) many of them.  Each such is to occur exactly
  \\ (n+1)! times, being ways the external decoration in L was, but which is
  \\ discarded by L_to_vpar().

  print1("checking ... ");
  for(n=0,4,
     my(all_L=random_decorated_all(n),
        all_vpar=apply(L_to_vpar,all_L));
     #Set(all_L)    == #all_L                   || error(); \\ all L distinct
     #Set(all_vpar) == vpar_count_ordered(n,1)  || error(); \\ ordered forests

     for(i=1,#all_vpar,
        my(vpar=all_vpar[i],
           L=all_L[i]);
        vpar_is_preorder(vpar)                              || error();
        vector_count_occurrences(all_vpar, vpar) == (n+1)!  || error();
        L_to_vpar_by_traverse(L) == vpar                    || error();
     );

     if(n==2,
        \\ The example L shown above is generated, and just once.
        vector_count_occurrences(all_L, [4, 3,1, 2,5]) == 1  || error());
     );
  print("good");
}

\\-----------------------------------------------------------------------------
